#!/usr/bin/env python
#
# Script for monitoring and freeing disk space.
#
# Delete files in specified directories according to low and high
# watermarks of the number of files allowed in the directory and low
# and high watermarks for the amount of space available on the
# filesystems containing those directories.  Files are deleted from
# oldest to newest within each directory until below the desired file
# counts and then within the filesystem as a whole until below the
# desired disk usage.  Files that are open by other processes are left
# in place.  Any invalid symlinks remaining in the directories after
# file deletion are removed.
#
# The script is written to check whether it needs to start freeing up
# space with minimal impact so that it can be run frequently
# (e.g. from a cron job).  It should not use any significant processor
# and disk bandwidth resources unless it needs to free up space and
# even then it is written to minimize impact on the system.

import datetime
import errno
import fcntl
import getopt
import logging
import logging.handlers
import os
import select
import string
import subprocess
import sys
import time

LOGGER_NAME = "freespace"
LOG = logging.getLogger(LOGGER_NAME)

DEFAULT_PIDFILE = "/var/run/freespace.py.pid"
DEFAULT_STOPFILE = "/var/lock/freespace.stop"
PIDFILE_SLEEP = 5

# NOTE ON LOGGING: Information messages should only be logged by
# free_up_space_in_filesystem() and its descendants in the call graph
# to prevent filling up syslog with useless periodic log messages.
# However, errors should be logged elsewhere if they occur.  Debug
# logging may be used anywhere since it isn't enabled by default.

def usage():
    print(("""\
Usage: freespace_monitor.py [options]...

Options are processed in sequential order and may be any of:

   -r,--runlevel configuration:
      Specifies up to which level of freespace removal should happen.  The
      default is level 0.  If a level is specified, each level is
      evaluated in order to the level specified.

   -l,--level configuration:
      Specifies a level for all subsequent command line arguments.  This can
      be used to add more entries (filesystem, directory, dirtree) to a given
      level that may or may not be specified in the config file.  Levels
      should increase by 1.

   -f,--filesystem <path>:<usage_low>:<usage_high>:<min_bytes>:<min_files>
      Add a filesystem to be monitored along with related configuration
      as colon separated fields.  The fields are:
          <path>: Full mount path for the filesystem
          <usage_low>: Target disk usage %% to reach when freeing space
          <usage_high>: Disk usage %% to trigger attempt to free space
          <min_bytes>: Minimum free bytes to trigger attempt to free space
          <min_files>: Minimum free files to trigger attempt to free space
      If the provided path does not correspond to a filesystem mount
      point, the other configured filesystyems will be examined, and if
      one is the parent filesystem, the requirements of the two
      filesystems will be merged.

   -d,--directory <path>:<files_low>:<files_high>
      Add a directory to be monitored long with related configuration as
      colon separated fields.  The fields are:
          <path>: Full directory path to the directory
          <files_low>: Target file count to reach when freeing space
          <files_high>: File count to trigger attempt to free space

   -p,--prefix <prefix>:<days>
      Add a prefix of files to be monitored and deleted. The fields are:
          <prefix>: Prefix of files
          <days>: Any files that matches the prefix and older than days
                  are deleted

   -t,--dirtree <path>:<files_low>:<files_high>
      Add a directory tree to be monitored long with related configuration as
      colon separated fields.  The fields are:
          <path>: Full directory path to the directory
          <files_low>: Target file count to reach when freeing space
          <files_high>: File count to trigger attempt to free space

      Note that the watermarks for dirtree should be set higher than
      they generally are for regular directory configuration as it will
      include all of the files in the subdirectories too.

   -c,--config <path>
      Read filesystem and/or directory configuration from the file
      specified by <path>

   -v,--verbose
      Log information about progress to both standard output and
      syslog.  If this flag is *not* specified no output will be
      generated and only error and warning level log messages will be
      logged to syslog.

   -w,--wait <timeout>
      Only one freespace process should be active at a time.  Normally
      if a second process is invoked an error is printed and that
      process exits with an error.  If this option is specified it
      will wait until the existing process completes and then run.  If
      the specified timeout (in seconds) is exceeded it will instead
      exit with an error.  A timeout value of zero means wait forever.
      A negative value means exit immediately (equivalent to not
      specifying the wait option).

   --pidfile
      Path to the process id (PID) file that is updated while running.
      The default is '%s'.  The PID file is
      used to ensure that only one instance runs at a time and must be
      writable.

   -h,--help
      Print this information.

   --debug
      Log extra debugging information.  By default this will go to
      syslog but if the verbose option is specified it will go to
      standard output as well.

Where configuration from options conflict, the last specified on the
command line is used.  Monitoring a filesystem without specifying at
least one directory on that filesystem will only log warnings about
out of space conditions.  Attempting to monitor a directory that is
not within a specified filesystem will generate an error since all
processing is driven based on space needs each filesystem.

The exit code will be 0 unless enough free space could not be freed in
which case it will be 1.  In the latter case, information on why space
could not be freed will be present in syslog.
"""  % (DEFAULT_PIDFILE,)))

OPTIONS = "e:r:l:f:d:p:t:c:w:vh"
LONG_OPTIONS = [
    "exec=",
    "runlevel=",
    "level=",
    "filesystem=",
    "directory=",
    "prefix=",
    "dirtree=",
    "config=",
    "wait=",
    "verbose",
    "pidfile=",
    "help",
    "debug"
]

class UsageException(Exception):
    "Raised when arguments or config are incorrect"
    pass

class SingletonException(Exception):
    "Raised when another process is already running."
    pass

class YieldException(Exception):
    "Raised when another process is running exclusively."
    pass

def main():
    options = {}
    failed = 0
    pidfd = None
    try:
        config, options, levels = getConfig()
        pidfd = lock_pidfile(options["pidfile"], options["wait"])
        if not freespace(config, options, levels):
            failed = 1
    except UsageException as e:
        print(("%s\n" % str(e)))
        LOG.error("Invalid usage: %s", repr(str(e)))
        usage()
        failed = 1
    except SingletonException as e:
        LOG.warn(str(e))
        failed = 1
    except YieldException as e:
        LOG.warn(str(e))
        failed = 1
    except Exception as e:
        LOG.exception(e)
        failed = 1
    finally:
        try:
            free_pidfile(pidfd, options.get("pidfile", None))
        except Exception as e:
            LOG.error("Exception occurred while freeing pidfile")
            LOG.exception(e)
    return failed

def get_mount_point(path):
    # Given a filesystem path, return the mount point
    # of the filesystem on which the path resides.
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path

def coalesce(levels, fsrecs):
    # Coalesce fsrecs that would operate on the same filesystem into a
    # single fsrec.
    new_fsrecs = {}
    for level in levels:
        # Group all fsrecs by the mount point of their containing filesystem
        mount_point_map = {}
        new_fsrecs[level] = []
        for fsrec in fsrecs.get(level, []):
            mount_point = get_mount_point(fsrec['filesystem_path'])
            if mount_point not in mount_point_map:
                mount_point_map[mount_point] = []
            mount_point_map[mount_point].append(fsrec)

        # Now coalesce them. Use the most aggresive high/low watermarks of
        # all the fsrecs found, and sum the minimum free byte and file
        # counts.
        for mount_point in list(mount_point_map.keys()):
            low_wat = high_wat = 100
            file_cnt = byte_cnt = 0
            mp_fsrecs = mount_point_map[mount_point]
            if len(mp_fsrecs) > 1:
                LOG.debug("Coalescing %d filesystem records into mount point "
                          "%s: %r" % (len(mp_fsrecs), mount_point, mp_fsrecs))
            for fsrec in mp_fsrecs:
                low_wat = min(low_wat, fsrec['low_space_usage_watermark'])
                high_wat = min(high_wat, fsrec['high_space_usage_watermark'])
                file_cnt += fsrec['minimum_free_file_count']
                byte_cnt += fsrec['minimum_free_bytes']
            new_fsrec = {
                'filesystem_path': mount_point,
                'low_space_usage_watermark': low_wat,
                'high_space_usage_watermark': high_wat,
                'minimum_free_file_count': file_cnt,
                'minimum_free_bytes': byte_cnt,
                'directories': [],
                'dirtrees': [],
                'prefixes': [],
            }
            new_fsrecs[level].append(new_fsrec)
    return new_fsrecs

def getConfig():
    # The configuration is represented as a dictionary keyed by
    # filesystem mount path where the values are another dictionary
    # with the keys:
    #    high_space_usage_watermark, low_space_usage_watermark,
    #    minimum_free_bytes, minimum_free_file_count, and directories
    # Values for all but the last are numerical.  The directories key
    # contains a list of three element tuples where the elements are
    # the directory, low_file_count_watermark and high_file_count_watermark.

    # Setup default logging to syslog at error level to log option
    # processing errors.  This may get further modified later based on
    # options processing.
    log = logging.getLogger(LOGGER_NAME)
    syslogHandler = LocalSysLogHandler()
    syslogFormatter = logging.Formatter("%(levelname)s:%(message)s")
    syslogHandler.setFormatter(syslogFormatter)
    loglevel = logging.WARNING
    syslogHandler.setLevel(loglevel)
    log.addHandler(syslogHandler)
    log.setLevel(loglevel)

    verbose = False
    debug = False
    options = {
        "level": 0,
        "runlevel": 0,
        "pidfile" : DEFAULT_PIDFILE,
        "wait" : -1
    }
    config = {}

    execs = {}
    fsrecs = {}
    dirrecs = {}
    dtrecs = {}
    prefixrecs = {}

    execs[0] = []
    fsrecs[0] = []
    dirrecs[0] = []
    dtrecs[0] = []
    prefixrecs[0] = []

    try:
        parsed_options, parsed_arguments = getopt.getopt(sys.argv[1:],
                                                         OPTIONS,
                                                         LONG_OPTIONS)
    except getopt.GetoptError as e:
        raise UsageException("Invalid option: %s" % str(e))
    if len(parsed_arguments):
        raise UsageException("Command line must not have non-option"
                             + " arguments: %s" % parsed_arguments)
    current_level = 0
    for opt, arg in parsed_options:
        if opt in ["-r", "--runlevel"]:
            options["runlevel"] = parse_level(arg)
        elif opt in ["-l", "--level"]:
            current_level = options["level"] = parse_level(arg)
            if current_level not in execs:
                execs[current_level] = []
            if current_level not in fsrecs:
                fsrecs[current_level] = []
            if current_level not in dirrecs:
                dirrecs[current_level] = []
            if current_level not in dtrecs:
                dtrecs[current_level] = []
        elif opt in ["-f", "--filesystem"]:
            fsrecs[current_level].append(parse_filesystem_flds(arg,
              "command-line option"))
        elif opt in ["-d", "--directory"]:
            dirrecs[current_level].append(parse_directory_flds(arg,
              "command-line option"))
        elif opt in ["-p", "--prefix"]:
            prefixrecs[current_level].append(parse_prefix_flds(arg,
              "command-line option"))
        elif opt in ["-t", "--dirtree"]:
            dtrecs[current_level].append(parse_directory_flds(arg,
              "command-line option"))
        elif opt in ["-e", "--exec"]:
            execs[current_level].append(parse_exec_flds(arg,
              "command-line option"))
        elif opt in ["-c", "--config"]:
            read_config_file(arg, execs, fsrecs, dirrecs, dtrecs, prefixrecs)
        elif opt in ["-w", "--wait"]:
            try:
                options["wait"] = int(arg)
            except ValueError as e:
                raise UsageException("Invalid wait period '%s'." % arg
                                     +  " Must be an integer")
        elif opt in ["--pidfile"]:
            options["pidfile"] = arg
        elif opt in ["-h", "--help"]:
            usage()
            sys.exit(0)
        elif opt in ["-v", "--verbose"]:
            verbose = True
        elif opt in ["--debug"]:
            debug = True
            log.warn("Debug logging enabled")
        else:
            raise Exception("Unhandled option: %s" % opt)

    if debug:   # Select appropriate logging level
        loglevel = logging.DEBUG
    elif verbose:
        loglevel = logging.INFO
    if verbose: # Setup stdout logging handler if verbose
        stdoutHandler = logging.StreamHandler(sys.stdout)
        stdoutFormatter = logging.Formatter("%(levelname)s: %(message)s")
        stdoutHandler.setFormatter(stdoutFormatter)
        stdoutHandler.setLevel(loglevel)
        log.addHandler(stdoutHandler)
    syslogHandler.setLevel(loglevel)
    log.setLevel(loglevel)

    s = set(fsrecs.keys())
    s.update(list(dirrecs.keys()))
    s.update(list(dtrecs.keys()))
    s.update(list(prefixrecs.keys()))
    levels = list(s)

    fsrecs = coalesce(levels, fsrecs)

    for level in levels:
        for fsrec in fsrecs.get(level, []):
            if level not in config:
                config[level] = {}
            if level in execs:
                # Add any command exececutions that need to happen to
                # the fsconfig
                for e in execs[level]:
                    if get_mount_point(e["fspath"]) == fsrec["filesystem_path"]:
                        key = "exec_%s" % e["phase"]
                        if key not in fsrec:
                            fsrec[key] = []
                        fsrec[key].append((e["cmd"], e["args"]))
                        LOG.debug("Added exec: %s to fsrec: %s" % (e, fsrec))
            config[level][fsrec["filesystem_path"]] = fsrec
        if level not in config:
            raise UsageException(("No filesystem specified for level: %d" % level))
        fspaths = list(config[level].keys())
        for dirrec in dirrecs.get(level, []):
            fspath = lpm(dirrec[0], fspaths)
            if not fspath:
                raise UsageException("Could not identify filesystem for"
                                     + " directory '%s'" % dirrec[0])
            config_add_or_replace_dirrec(dirrec,
                                         config[level][fspath]["directories"])
        for dtrec in dtrecs.get(level, []):
            fspath = lpm(dtrec[0], fspaths)
            if not fspath:
                raise UsageException("Could not identify filesystem for"
                                     + " directory '%s'" % dtrec[0])
            config_add_or_replace_dirrec(dtrec,
                                         config[level][fspath]["dirtrees"])
        for fspath in fspaths:
            for prefixrec in prefixrecs.get(level, []):
                config_add_or_replace_prefixrec(
                    prefixrec, config[level][fspath]["prefixes"])

    return config, options, levels

def check_stop_file(stop_file):
    "Check for stop file"
    if os.path.exists(stop_file):
        delete_stop_file = False
        try:
            fh = open(stop_file, 'r')
            stime = fh.readline()
            fh.close()
            start_time = time.mktime(time.strptime(stime.strip(),
                                     "%Y-%m-%d %H:%M:%S"))
            run_hour = (time.time() - start_time) / 3600
            if run_hour > 1:
                LOG.warn("Stop file '%s' exists for %.1f hours"
                         % (stop_file, run_hour))
                delete_stop_file = True
        except Exception as e:
            LOG.error("Cannot read stop file '%s': %s"
                      % (stop_file, str(e)))
            delete_stop_file = True
        if delete_stop_file:
            LOG.warn("Delete obsolete stop file '%s'" % stop_file)
            try:
                os.remove(stop_file)
            except Exception as e:
                LOG.error("Cannot remove stop file '%s': %s"
                          % (stop_file, str(e)))
        else:
            # Yield if other process like upgrading is running
            raise YieldException("Yield to other process like upgrading, "
                     "which has been running for %.1f hour. "
                     "If you do want to run and are aware of the consequence, "
                     "please remove the stop file '%s'."
                     % (run_hour, stop_file))

def read_config_file(path, execs, fsrecs, dirrecs, dtrecs, prefixrecs):
    # fsrecs and dirrecs are appended to in place
    f = None
    current_level = 0
    try:
        f = open(path, "rt")
        for i, l in enumerate(f):
            l = l.strip()
            if not len(l) or l[0] == "#":
                # Blank line or comment
                continue
            source = "configuration file '%s' line %d" % (path, i+1)
            if l.startswith("level:"):
                current_level = parse_level(l[6:])
                LOG.debug("Current level: %d" % current_level)
                if current_level not in fsrecs:
                    fsrecs[current_level] = []
                if current_level not in dirrecs:
                    dirrecs[current_level] = []
                if current_level not in dtrecs:
                    dtrecs[current_level] = []
                if current_level not in execs:
                    execs[current_level] = []
            elif l.startswith("exec:"):
                execs[current_level].append(parse_exec_flds(l[5:], source))
            elif l.startswith("filesystem:"):
                fsrecs[current_level].append(parse_filesystem_flds(l[11:],
                                                                   source))
            elif l.startswith("directory:"):
                dirrecs[current_level].append(parse_directory_flds(l[10:],
                                                                   source))
            elif l.startswith("dirtree:"):
                dtrecs[current_level].append(parse_directory_flds(l[8:],
                                                                  source))
            elif l.startswith("prefix:"):
                prefixrecs[current_level].append(parse_prefix_flds(l[7:],
                                                                   source))
            else:
                raise UsageException(("Invalid directive in %s") % (source))
    finally:
        if f:
            f.close()

def parse_level(text):
    val = 0
    try:
        val = int(text)
    except ValueError:
        raise UsageException(("Level must be an integer"))
    if val < 0:
        raise UsageException(("Level must be a positive integer"))
    return val

def parse_exec_flds(text, source):
    rec = {}
    flds = text.split(":")
    if len(flds) != 3:
        raise UsageException(("Invalid number of fields in %s exec"
                              + " configuration: %s") % (source, text))
    phase = flds[0].strip()
    if not phase in ["before", "during", "after"]:
        raise UsageException(("Phase must be one of \"before, during, after\""
                              + " in exec entry: %s" % text))
    rec["phase"] = phase
    fspath = flds[1].strip()
    if not os.path.isdir(fspath):
        raise UsageException("Directory '%s' does not exist" %
                             fspath)
    rec["fspath"] = fspath
    cmd = flds[2].strip()
    if " " in cmd:
        cmd, args = cmd.split(" ", 1)
    else:
        args = ""
    if not os.path.exists(cmd):
        raise UsageException("Command '%s' does not exist" % cmd)
    rec["cmd"] = cmd
    rec["args"] = args
    return rec

def parse_filesystem_flds(text, source):
    rec = {}
    flds = text.split(":")
    if len(flds) != 5:
        raise UsageException(("Invalid number of fields in %s filesystem"
                              + " configuration: %s") % (source, text))
    fspath = flds[0].strip()
    if not os.path.isdir(fspath):
        raise UsageException("Directory '%s' does not exist" %
                             fspath)
    rec["filesystem_path"] = fspath
    used_space_low = flds[1].strip()
    try:
        used_space_low = int(used_space_low)
    except ValueError as e:
        raise UsageException(("Space usage low watermark ('%s') is not an"
                              + " integer value in %s filesystem"
                              + " configuration: %s")
                             % (used_space_low, source, text))
    if used_space_low < 0 or used_space_low > 100:
        raise UsageException(("Space usage low watermark ('%d') is not between"
                              + " 0 and 100 percent in %s filesystem"
                              + " configuration: %s")
                             % (used_space_low, source, text))
    rec["low_space_usage_watermark"] = used_space_low
    used_space_high = flds[2].strip()
    try:
        used_space_high = int(used_space_high)
    except ValueError as e:
        raise UsageException(("Space usage high watermark ('%s') is not an"
                              + " integer value in %s filesystem"
                              + " configuration: %s")
                             % (used_space_high, source, text))
    if used_space_high < 0 or used_space_high > 100:
        raise UsageException(("Space usage high watermark ('%d') is not "
                              + " between 0 and 100 percent in %s filesystem"
                              + " configuration: %s")
                             % (used_space_high, source, text))
    if used_space_high < used_space_low:
        raise UsageException(("Space usage high watermark ('%d') less than"
                              + " low watermark ('%d') in %s filesystem"
                              + " configuration: %s")
                             % (used_space_high, used_space_low, source, text))
    rec["high_space_usage_watermark"] = used_space_high
    minimum_free_bytes = flds[3].strip()
    try:
        minimum_free_bytes = int(minimum_free_bytes)
    except ValueError as e:
        raise UsageException(("Minimum free bytes value ('%s') is not an"
                              + " integer value in %s filesystem"
                              + " configuration: %s")
                             % (minimum_free_bytes, source, text))
    if minimum_free_bytes < 0:
        raise UsageException(("Minimum free bytes value ('%d') is negative "
                              + " in %s filesystem configuration: %s")
                             % (minimum_free_bytes, source, text))
    rec["minimum_free_bytes"] = minimum_free_bytes
    minimum_free_files = flds[4].strip()
    try:
        minimum_free_files = int(minimum_free_files)
    except ValueError as e:
        raise UsageException(("Minimum free files value ('%s') is not an"
                              + " integer value in %s filesystem"
                              + " configuration: %s")
                             % (minimum_free_files, source, text))
    if minimum_free_files < 0:
        raise UsageException(("Minimum free files value ('%d') is negative "
                              + " in %s filesystem configuration: %s")
                             % (minimum_free_files, source, text))
    rec["minimum_free_file_count"] = minimum_free_files
    rec["directories"] = []
    rec["dirtrees"] = []
    return rec

def parse_directory_flds(text, source):
    rec = []
    flds = text.split(":")
    if len(flds) != 3:
        raise UsageException(("Invalid number of fields in %s directory"
                              + " configuration: %s") % (source, text))
    dirpath = flds[0].strip()
    rec.append(dirpath)
    file_count_low = flds[1].strip()
    try:
        file_count_low = int(file_count_low)
    except ValueError as e:
        raise UsageException(("File count low watermark ('%s') is not an"
                              + " integer value in %s directory"
                              + " configuration: %s")
                             % (file_count_low, source, text))
    if file_count_low < 0:
        raise UsageException(("File count low watermark value ('%d') is"
                              + " negative in %s directory configuration: %s")
                             % (file_count_low, source, text))
    rec.append(file_count_low)
    file_count_high = flds[1].strip()
    try:
        file_count_high = int(file_count_high)
    except ValueError as e:
        raise UsageException(("File count high watermark ('%s') is not an"
                              + " integer value in %s directory"
                              + " configuration: %s")
                             % (file_count_high, source, text))
    if file_count_high < 0:
        raise UsageException(("File count high watermark value ('%d') is"
                              + " negative in %s directory configuration: %s")
                             % (file_count_high, source, text))
    if file_count_high < file_count_low:
        raise UsageException(("File count high watermark ('%d') less than"
                              + " low watermark ('%d') in %s directory"
                              + " configuration: %s")
                             % (file_count_high, file_count_low, source, text))
    rec.append(file_count_high)
    return rec

def parse_prefix_flds(text, source):
    rec = []
    flds = text.split(":")
    if len(flds) != 2:
        raise UsageException(("Invalid number of fields in %s prefix"
                              + " configuration: %s") % (source, text))
    prefix = flds[0].strip()
    rec.append(prefix)
    days = flds[1].strip()
    try:
        days = int(days)
    except ValueError as e:
        raise UsageException(("Days ('%s') is not an"
                              + " integer value in %s prefix"
                              + " configuration: %s")
                             % (days, source, text))
    if days < 0:
        raise UsageException(("Days value ('%d') is"
                              + " negative in %s prefix configuration: %s")
                             % (days, source, text))
    rec.append(days)
    return rec

def lpm(text, prefixes):
    best_match = None
    for p in prefixes:
        if text.startswith(p):
            if not best_match:
                best_match = p
            elif len(best_match) < len(p):
                best_match = p
    return best_match

def config_add_or_replace_dirrec(dirrec, dirrecs):
    # Modifies dirrecs in place
    for i, d in enumerate(dirrecs):
        if d[0] == dirrec[0]:
            dirrecs[1] = dirrec
            return
    dirrecs.append(dirrec)

def config_add_or_replace_prefixrec(prefixrec, prefixrecs):
    # Modifies prefixrecs in place
    for i, d in enumerate(prefixrecs):
        if d[0] == prefixrec[0]:
            prefixrecs[1] = prefixrec
            return
    prefixrecs.append(prefixrec)

def lock_pidfile(fn, wait):
    LOG.debug("Locking pidfile: %s", fn)
    # Don't attempt to create pidfile if already out of space in that
    # partition.
    piddir = os.path.dirname(fn)
    try:
        percent_usage, bytes_free, files_free = get_disk_usage(piddir)
    except OSError as e:
        if e.errno == errno.ENOENT:
            LOG.error("PID file directory does not exist: %s", piddir)
            LOG.warn("Attempting to free space without creating PID file")
            return None
        else:
            raise e
    if bytes_free < 100 or (files_free is not None and files_free < 5):
        # Thresholds deliberately slightly greater than 0 to mitigate races
        LOG.warn("No space to create PID file in dir: %s", piddir)
        LOG.warn("Attempting to free space without creating PID file")
        return None
    fd = update_pidfile(fn)
    elapsed = 0
    while not fd and (wait == 0 or elapsed < wait):
        LOG.warn("Waiting for already running process to exit.")
        time.sleep(PIDFILE_SLEEP)
        elapsed = elapsed + PIDFILE_SLEEP
        fd = update_pidfile(fn)
    if not fd:
        if wait < 0:
            raise SingletonException("Another process is already running.")
        else:
            raise SingletonException("Timed out waiting for"
                                     + " another process to exit.")
    return fd

def update_pidfile(fn):
    flags = os.O_WRONLY | os.O_EXCL | os.O_CREAT
    fd = None
    while not fd:
        try:
            fd = os.open(fn, flags)
        except OSError as e:
            LOG.debug("Couldn't open pidfile")
            fd = None
            if e.errno == errno.EEXIST:
                LOG.debug("The pidfile exists and may be orphaned")
                if not remove_orphan_pidfile(fn):
                    return None   # Still locked by someone else
            else:
                raise e   # Unexpected exception
    # Exceptions deliberately not caught
    fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    try:
        os.write(fd, str(os.getpid()) + "\n")
    except TypeError:
        os.write(fd, str.encode(str(os.getpid()) + "\n"))
    return fd

def remove_orphan_pidfile(fn):
    try:
        fd = os.open(fn, os.O_RDONLY)
        fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        os.remove(fn)
        os.close(fd)
    except (OSError, IOError) as e:
        return False
    LOG.debug("Removed orphaned pidfile")
    return True

def free_pidfile(fd, fn):
    # Exceptions deliberately not caught
    if fd:
        # Should only get here if succeeded in creating and locking pidfile.
        os.remove(fn)
        os.close(fd)
        LOG.debug("Freed pidfile")

def freespace(config, options, levels):
    num_failures = 0
    LOG.debug("Level specified: %d" % options["runlevel"])
    # Yield if other process like upgrading is running
    check_stop_file(DEFAULT_STOPFILE)
    for level in levels:
        if level > options["runlevel"]:
            break
        LOG.info("Initiating level %d cleanup" % level)
        for fspath, fsconfig in list(config[level].items()):
            succeeded = True
            if not os.path.isdir(fspath):
                LOG.error("Filesystem '%s' mount point does not exist", fspath)
                continue
            used_space_high = fsconfig["high_space_usage_watermark"]
            minimum_bytes_free = fsconfig["minimum_free_bytes"]
            minimum_free_file_count = fsconfig["minimum_free_file_count"]
            # If all threshholds are zero, this is the CLI doing a
            # "clear disk-space fully" and since it can never succeed
            # in reducing disk usage to zero, don't treat that as
            # a failure.
            reclaim_fully_request = used_space_high == minimum_bytes_free == \
                                    minimum_free_file_count == 0
            percent_usage, bytes_free, files_free = get_disk_usage(fspath)

            # Remove any files that matches the given prefix in fsconfig
            # that are older then the specified config.
            remove_old_files_in_filesystem(fspath, fsconfig)

            if (percent_usage > used_space_high
                or bytes_free < minimum_bytes_free
                or (files_free is not None and
                    files_free < minimum_free_file_count)):
                succeeded = free_up_space_in_filesystem(fspath, fsconfig)
            else:
                LOG.info("Filesystem '%s' has enough space", fspath)
                LOG.info("Filesystem '%s', usage is %d%%,"
                         + " free bytes is %d, and free file count is %s",
                         fspath, percent_usage, bytes_free,
                         "(not relevant for this filesytem)"
                         if files_free is None else files_free)
            if not succeeded and not reclaim_fully_request:
                num_failures += 1
                files_free_str = ("(not relevant for this filesystem)"
                                  if files_free is None else "%d" % files_free)
                LOG.warning("Unable to reclaim sufficient free space or free "
                            "files on filesystem '%s'", fspath)
                LOG.warning("Filesystem '%s', usage is %d%%,"
                         + " free bytes is %d, and free file count is %s",
                         fspath, percent_usage, bytes_free, files_free_str)
    return num_failures == 0

def get_disk_usage(directory):
    # Assumes directory path exists.  If not this will throw an
    # exception. Returns a tuple: (percentage_used, absolute_free_bytes,
    # free_files). If the filesystem type does not have limits on the
    # number of files (e.g. btrfs), then free_files will be None.
    s = os.statvfs(directory)
    absolute_free_bytes = s.f_frsize * s.f_bfree
    percentage_used = ((100 * (s.f_blocks - s.f_bavail)) / s.f_blocks)
    # If the filesystem reports f_files == 0, then we assume it
    # doesn't have a limit.
    free_files = None if s.f_files == 0 else s.f_ffree
    return (percentage_used, absolute_free_bytes, free_files)

def enough_free(fspath, fsconfig, phase):
    used_space_high = fsconfig["high_space_usage_watermark"]
    used_space_low = fsconfig["low_space_usage_watermark"]
    minimum_bytes_free = fsconfig["minimum_free_bytes"]
    minimum_free_file_count = fsconfig["minimum_free_file_count"]
    percent_usage, bytes_free, files_free = get_disk_usage(fspath)
    done = True
    if used_space_low > 0 and percent_usage > used_space_low:
        LOG.warn("[phase: %s] filesystem '%s'"
                 + " , usage (%d%%) is still above low watermark (%d%%)",
                 phase, fspath, percent_usage, used_space_low)
        done = False
    if minimum_bytes_free > 0 and bytes_free < minimum_bytes_free:
        LOG.warn("[phase: %s] filesystem '%s',"
                 + " free space (%d) is still below minimum free space (%d)",
                 phase, fspath, bytes_free, minimum_bytes_free)
        done = False
    if (minimum_free_file_count > 0 and files_free is not None and
        files_free < minimum_free_file_count):
        LOG.warn("[phase: %s] filesystem '%s',"
                 + " free file count (%d) is still below minimum free"
                 + " count (%d)",
                 phase, fspath, files_free, minimum_free_file_count)
        done = False
    files_free_str = ("(not relevant for this filesystem)" if files_free is
                     None else "%d" % files_free)
    LOG.info("[phase: %s] filesystem '%s',"
             + " usage is %d%%, free bytes is %d, and free file count is %s",
             phase, fspath, percent_usage, bytes_free, files_free_str)
    return done

def free_up_space_in_filesystem(fspath, fsconfig):
    cmds = fsconfig.get("exec_before", [])
    for c in cmds:
        run_command(c)
        # Check after every command run
        if enough_free(fspath, fsconfig, "after 'before' cmd: %s" % str(c)):
            return True
    # And check before we move on
    if enough_free(fspath, fsconfig, "after all 'before' commands"):
        return True
    LOG.warn("Freeing space in filesystem '%s'", fspath)
    dir_filelists = []
    for dirpath, file_count_low, file_count_high in fsconfig["directories"]:
        if not os.path.isdir(dirpath):
            LOG.warn("Monitored directory '%s' does not exist", dirpath)
            continue
        LOG.info("Examining directory '%s'", dirpath)
        files = get_files(dirpath)
        numfiles = len(files)
        if numfiles > file_count_high:
            LOG.warn("File count high watermark %d exceeded for"
                     + " directory '%s'",
                     file_count_high, dirpath)
            free_until_below_file_count(dirpath, files, file_count_low)
        dir_filelists.append(files)
    for dirpath, file_count_low, file_count_high in fsconfig["dirtrees"]:
        if not os.path.isdir(dirpath):
            LOG.warn("Monitored directory '%s' does not exist", dirpath)
            continue
        LOG.info("Examining directory tree '%s'", dirpath)
        files, directories = get_dirtree(dirpath)
        numfiles = len(files)
        if numfiles > file_count_high:
            LOG.warn("File count high watermark %d exceeded for"
                     + " directory '%s'",
                     file_count_high, dirpath)
            free_until_below_file_count(dirpath, files, file_count_low)
        dir_filelists.append(files)
        for d in directories:
            remove_directory_if_empty(d)
    # Quick check before doing any of the "during" commands
    if enough_free(fspath, fsconfig, "after directory size reduction"):
        return True
    cmds = fsconfig.get("exec_during", [])
    for c in cmds:
        run_command(c)
        # Check after each "during" command run
        if enough_free(fspath, fsconfig, "after 'during' cmd: %s" % str(c)):
            return True
    remaining_files = merge_filelists(dir_filelists)
    used_space_low = fsconfig["low_space_usage_watermark"]
    minimum_bytes_free = fsconfig["minimum_free_bytes"]
    if not enough_free(fspath, fsconfig, "after all 'during' commands"):
        free_until_enough_space(fspath, remaining_files,
                                used_space_low, minimum_bytes_free)
    for dirpath, file_count_low, file_count_high in fsconfig["directories"]:
        remove_invalid_symlinks(dirpath)
    if enough_free(fspath, fsconfig, "after global age-based removal"):
        return True
    cmds = fsconfig.get("exec_after", [])
    for c in cmds:
        run_command(c)
        # Check after every command is run
        if enough_free(fspath, fsconfig, "after 'after' cmd: %s" % str(c)):
            return True
    # And do a final check
    if enough_free(fspath, fsconfig, "final check"):
        return True
    return False

def remove_old_files_in_filesystem(fspath, fsconfig):
    if "prefixes" not in fsconfig:
        return
    # Refresh for the open file cache for the entire filesystem
    # before final deletion phase
    open_file_cache = create_open_file_cache(fspath)

    for dirpath, _, _ in fsconfig["directories"]:
        if not os.path.isdir(dirpath):
            LOG.warn("Monitored directory '%s' does not exist", dirpath)
            continue
        LOG.info("Examining directory '%s'", dirpath)
        files = get_files(dirpath)
        if not len(files):
            continue
        for curr_file in files:
            fn, mtime = curr_file
            is_older_file = False
            for prefix, days in fsconfig["prefixes"]:
                if mtime < time.time() - (86400 * days):
                    is_older_file = True
                    if prefix in fn:
                        remove_file(fn, mtime, open_file_cache)
                        break
            # The file is newer than any of the prefix day filter.
            # Break out of the loop, since succeeding files in the list
            # is guaranteed to be newer.
            if not is_older_file:
                break

def run_command(c):
    # Execute a command capturing stdout and stderr.  stdout will be logged at
    # level INFO and stderr at ERR.
    LOG.info("Executing command: %s (args: %s)" % (c[0],
      " ".join([x for x in c[1:]])))
    pobj = subprocess.Popen(c, stdout=subprocess.PIPE,
      stderr=subprocess.PIPE)
    def _print_complete_lines(text, error=False):
        while True:
            if "\n" in text:
                pos = text.find("\n")
                line = text[:pos].strip()
                if error:
                    LOG.error(line)
                else:
                    LOG.info(line)
                # Chop off line
                text = text[pos+1:]
            else:
                break
        return text
    def _process_data(stdout, stderr, pobj):
        rlst, wlst, xlst = select.select([pobj.stdout, pobj.stderr],
          [], [pobj.stdout], 1)
        data = ""
        if len(rlst) > 0:
            for x in rlst:
                data = os.read(x.fileno(), 256)
                if x.fileno() == pobj.stdout.fileno():
                    stdout += data
                    stdout = _print_complete_lines(stdout)
                else:
                    stderr += data
                    stderr = _print_complete_lines(stderr, error=True)
        return stdout, stderr
    out, err = "", ""
    # Keep processing data until the process finishes
    while pobj.poll() == None:
        out, err = _process_data(out, err, pobj)
    # Process any remaining data
    while True:
        out, err = _process_data(out, err, pobj)
        # Print any remaining data
        if len(out) > 0:
            LOG.info(out)
        if len(err) > 0:
            LOG.error(err)
        if len(out) == 0 and len(err) == 0:
            break
    LOG.info("Command \"%s\", Exit status: %d" % (c, pobj.returncode))
    return pobj.returncode

def get_dirtree(directory):
    # Get files in directory, filtering symlinks.  This returns a tuple of
    # lists.  The first list is of tuples of (file, mtime) and the second list
    # is just a listing of the directories in the subtree.
    paths = []
    dirs = []
    if not os.path.isdir(directory):
        LOG.error("get_dirtree called on non-directory: %s" % directory)
        return paths, dirs
    for root, directories, files in os.walk(directory):
        for d in directories:
            dirs.append(os.path.join(root, d))
        for f in files:
            fname = os.path.join(root, f)
            if not os.path.islink(fname):
                paths.append((fname, os.path.getmtime(fname)))
    paths.sort(key=lambda x: x[1])
    return paths, dirs

def get_files(directory):
    # Get files in directory, filtering symlinks and subdirectories.
    # If subdirectories need to be cleared they should be included in
    # the list of directories to be examined.  The returned list
    # contains items of two elements, full path to the file and the
    # modification time of the file.
    paths = []
    if os.path.isdir(directory):
        for fn in os.listdir(directory):
            p = os.path.join(directory, fn)
            if not (os.path.islink(p) or os.path.isdir(p)):
                paths.append((p, os.path.getmtime(p)))
    paths.sort(key=lambda x: x[1])
    return paths

def free_until_below_file_count(directory, files, files_count_target):
    # Modifies the list of files in the files argument in place
    # Cache open files for performance reasons
    open_file_cache = create_open_file_cache(directory)
    skipped, removed = 0, 0
    while len(files) and len(files) + skipped > files_count_target:
        fn, mtime = files.pop(0)
        if remove_file(fn, mtime, open_file_cache):
            removed += 1
        else:
            skipped += 1
    LOG.warn("Removed %d and skipped %d of the oldest files in directory '%s'",
             removed, skipped, directory)

def merge_filelists(dir_filelists):
    # The individual directory filelists are already in sorted order.
    # Create the merged list from them.  To minimize memory allocation
    # the entries are moved from the individual directory lists to the
    # new list that will be returned so that all the original lists
    # and the dir_filelists list will be empty after it completes.
    merged_files = []
    dir_filelists = [x for x in dir_filelists if len(x) > 0]
    while len(dir_filelists):
        # We could be smarter here since the list is always sorted
        # except for the last element that changed, but it is likely
        # to be so small that it probably isn't worth it.
        dir_filelists.sort(key=lambda x: x[0][1])
        merged_files.append(dir_filelists[0].pop(0))
        if not len(dir_filelists[0]):
            del(dir_filelists[0])
    return merged_files

def free_until_enough_space(fspath, files,
                            used_space_target, minimum_bytes_free):
    # Modifies the list of files in the files argument in place
    if not len(files):
        LOG.warn("Filesystem '%s' has no files in configured directories",
                 fspath)
        return False
    # Refresh for the open file cache for the entire filesystem
    # before final deletion phase
    open_file_cache = create_open_file_cache(fspath)
    percent_usage, bytes_free = get_disk_usage(fspath)[:2]
    skipped, removed = 0, 0
    while (len(files)
           and (percent_usage > used_space_target
                or bytes_free < minimum_bytes_free)):
        fn, mtime = files.pop(0)
        if remove_file(fn, mtime, open_file_cache):
            removed += 1
        else:
            skipped += 1
        percent_usage, bytes_free = get_disk_usage(fspath)[:2]
    LOG.warn(("Removed %d and skipped %d of the oldest files"
              + " in filesystem '%s'"), removed, skipped, fspath)

def remove_file(fn, mtime, open_file_cache):
    fd = None
    try:
        LOG.debug("Removing file '%s'", fn)
        # Test whether file is open somewhere else by trying to lock
        # exclusively.
        try:
            fd = os.open(fn, os.O_RDONLY)
            if fd < 0:
                LOG.warn("Could not open file '%s': %s", fn, os.strerror(fd))
                return False
        except (OSError, IOError) as e:
            LOG.warn("Could not open file '%s' for removal: %s", fn, str(e))
            return False
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError as e:
            LOG.warn("Could not lock file '%s' on fd %d: %s", fn, fd, str(e))
            return False
        opened = processes_with_file_open(fn, open_file_cache)
        if len(opened):
            LOG.warn("File '%s' is currently open by processes: %s",
                     fn, ", ".join([str(x) for x in opened]))
            return False
        if os.path.getmtime(fn) != mtime:
            LOG.warn("File '%s' modified after selected for removal, skipping",
                     fn)
            return False
        try:
            os.remove(fn)
        except Exception as e:
            LOG.warn("Could not remove file '%s': %s", fn, str(e))
            return False
        return True
    finally:
        if fd:
            os.close(fd)

def processes_with_file_open(fn, open_file_cache):
    # There are unavoidable races with using one of these methods to
    # determine whether the file is opened by any other processes
    # since there is no way to put a real lock on a file and guarantee
    # no one else can use it once obtained.
    #
    # ----
    # lsof implementation (very slow)
    # return len(os.popen("lsof -t '%s'" % fn).readlines())
    # ----
    # fuser implementation (slightly faster)
    # text = os.popen("fuser '%s' 2>&1" % fn).read()
    # return filter(lambda x: x, text[len(fn)+1:].strip().split(" "))
    # ----
    # Implementation by resolving links in /proc (4x faster than above)
    # mypid = os.getpid()
    # processes = []
    # for d in os.listdir("/proc"):
    #     try:
    #         pid = int(d)
    #     except ValueError, e:
    #         # Not a PID directory
    #         continue
    #     if pid == mypid:
    #         continue
    #     try:
    #         fds = os.listdir("/proc/%d/fd" % pid)
    #     except OSError, e:
    #         # Probably permission denied if not run as root.  In any
    #         # case, skip because we can't do anything.
    #         continue
    #     for l in fds:
    #         if os.path.realpath(l) == fn:
    #             processes.append(pid)
    # return processes
    # ----
    # Cached implementation: 700x faster, just less safe
    try:
        return open_file_cache[fn]
    except KeyError as e:
        return []

def remove_directory_if_empty(directory):
    if not os.path.exists(directory):
        return
    if len(os.listdir(directory)) == 0:
        LOG.debug("Removing empty directory: %s" % directory)
        os.rmdir(directory)

def remove_invalid_symlinks(directory):
    # Implementing with find is probably faster than in python, though
    # possibly less safe.
    if os.path.isdir(directory):
        os.system("find -L %s -maxdepth 1 -type l -delete" % directory)

def create_open_file_cache(directory):
    cache = {}
    print("PATH=/usr/bin:/usr/sbin lsof +D '%s' -x -l -n -P" % directory)
    text = os.popen(
        "PATH=/usr/bin:/usr/sbin lsof +D '%s' -x -l -n -P" % directory)
    try:
        next(text)   # Get rid of header line
    except TypeError:
        text = text.read().strip().split("\n")
        text = text[1:]
    except StopIteration:
        return cache
    for l in text:
        r = [x for x in l.split(" ") if x]
        pid = int(r[1])
        fn = r[8].strip()
        try:
            cache[fn].append(pid)
        except KeyError:
            cache[fn] = [pid]
    return cache

class LocalSysLogHandler(logging.handlers.SysLogHandler):
    # There are a number of problems with the standard Python library
    # SysLogHandler when working with many servers, starting with the
    # fact that it doesn't log the date in the required (rather
    # non-standard) format after the facility and level nor does it
    # ensure the inclusion of a <= 32 character tag before the message.
    #
    # This implementation wraps the standard library one to solve
    # these problems.  It formats the date as required and uses the
    # logger name as the tag, truncated if necessary.  If the
    # formatter used for the log messages includes a timestamp or the
    # logger name that information will be redundant although it may
    # provide more information such as a more complete timestamp.
    #
    # Because this application only needs local logging and the record
    # format is different for remote logging this implementation only
    # handles local logging.  A similar remote syslog handler could be
    # implemented if required.
    #
    # This implementation was informed by RFC3164:
    #     http://www.ietf.org/rfc/rfc3164.txt
    _months = {
        1 : 'Jan', 2 : 'Feb', 3 : 'Mar', 4 : 'Apr', 5: 'May', 6 : 'Jun',
        7 : 'Jul', 8 : 'Aug', 9 : 'Sep', 10 : 'Oct', 11: 'Nov', 12 : 'Dec'
    }

    def __init__(self, facility=logging.handlers.SysLogHandler.LOG_USER,
                 tag=None):
        logging.handlers.SysLogHandler.__init__(self, address="/dev/log",
                                                facility=facility)
        self.tag = tag
        if self.tag:
            self.tag = self._fmtSyslogTag(tag)

    def format(self, record):
        text = logging.handlers.SysLogHandler.format(self, record)
        ts = self.syslogTimestamp(record)
        tag = self.syslogTag(record)
        return "%s %s: %s" % (ts, tag, text)

    def syslogTimestamp(self, record):
        # Quoting from RFC3164:
        #
        # The TIMESTAMP field is the local time and is in the format of
        # "Mmm dd hh:mm:ss" (without the quote marks) where:
        #
        # Mmm is the English language abbreviation for the month of the
        # year with the first character in uppercase and the other two
        # characters in lowercase.  The following are the only acceptable
        # values:
        #
        # Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec
        #
        # dd is the day of the month.  If the day of the month is less
        # than 10, then it MUST be represented as a space and then the
        # number.  For example, the 7th day of August would be
        # represented as "Aug  7", with two spaces between the "g" and
        # the "7".
        #
        # hh:mm:ss is the local time.  The hour (hh) is represented in a
        # 24-hour format.  Valid entries are between 00 and 23,
        # inclusive.  The minute (mm) and second (ss) entries are between
        # 00 and 59 inclusive.
        dt = datetime.datetime.now()
        datespace = ''
        if dt.day < 10:
            datespace = ' '
        return "%s %s%d %02d:%02d:%02d" % (self._months[dt.month],
                                           datespace, dt.day,
                                           dt.hour, dt.minute, dt.second)

    def syslogTag(self, record):
        # The tag is a 32 character or less, preferably 7-bit ASCII
        # printable character name followed by a colon or right
        # bracket with no spaces or other non-printing characters
        if self.tag:
            return self.tag
        return self._fmtSyslogTag(record.name.split(".")[-1])

    def _fmtSyslogTag(self, text):
        tag = [x for x in text if x in string.ascii_letters]
        return "".join(tag[:31])

if __name__ == "__main__":
    sys.exit(main())
