# ***************************************************************************
# Copyright 2014-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
from copy import copy
import getpass
from itertools import permutations
import nsx_logging
import os
import re
import shlex
try:
    import simplejson
except ImportError:
    import json as simplejson
import sys
import time
import uuid


from cli.descriptors.cli_argument_descriptor import CliArgumentDescriptor
from cli.descriptors.cli_argument_descriptor import is_argument
from cli.descriptors.cli_argument_descriptor import resolve_argument_name
from cli.descriptors.cli_argument_descriptor import show_argument_description
from cli.descriptors.cli_command_descriptor import CliCommandDescriptor
from cli.utils.central_cli_api import async_invoke_and_wait
from cli.utils.central_cli_api import CentralCLIAPIException
from cli.utils.central_cli_utils import get_central_cli_nodes
from cli.utils.central_cli_utils import is_central_cli_command
from cli.utils.central_cli_utils import get_central_cli_user_input
from cli.utils.central_cli_utils import get_cmd_list_path_tuple
from cli.utils.constants import CENTRAL_CLI_START_TOKEN
from cli.utils.constants import CENTRAL_CLI_END_TOKEN
from cli.utils.constants import CENTRAL_CLI_STR
from cli.utils.constants import CMD_EXECUTED
from cli.utils.constants import CMD_EXECUTED_REQUESTED_EXIT
from cli.utils.constants import CMD_EXECUTED_WITH_ERROR_RESULT
from cli.utils.constants import CMD_MATCH_EXACT
from cli.utils.constants import CMD_MATCH_FAILED
from cli.utils.constants import CMD_MATCH_INCOMPLETE
from cli.utils.constants import CMD_MATCH_PARTIAL
from cli.utils.constants import CMD_MATCH_VALUE_MISMATCH
from cli.utils.constants import CMD_NOT_EXECUTED
from cli.utils.constants import END_TOKEN
from cli.utils.constants import EXECUTE_COMMAND_HELP_ID
from cli.utils.constants import MODE_ALL
from cli.utils.constants import MODE_ENABLE
from cli.utils.constants import MODE_CONFIGURE
from cli.utils.constants import MODE_INTERFACE
from cli.utils.constants import MODE_TIER0_SR
from cli.utils.constants import MODE_TIER1_SR
from cli.utils.constants import MODE_TIER0_VRF_SR
from cli.utils.constants import MODE_INTER_SR_VRF
from cli.utils.constants import MODE_REMOTE_TUNNEL_VRF
from cli.utils.constants import MODE_VRF
from cli.utils.constants import MODE_PATH
from cli.utils.constants import MODE_OFFLINEPATH
from cli.utils.constants import MODE_QUIT
from cli.utils.constants import OUTPUT_MODIFIER_COUNT
from cli.utils.constants import OUTPUT_MODIFIER_FIND
from cli.utils.constants import OUTPUT_MODIFIER_FIRST
from cli.utils.constants import OUTPUT_MODIFIER_IGNORE
from cli.utils.constants import OUTPUT_MODIFIER_JSON
from cli.utils.constants import OUTPUT_MODIFIER_LAST
from cli.utils.constants import OUTPUT_MODIFIER_MORE
from cli.utils.constants import OUTPUT_MODIFIER_SORT
from cli.utils.constants import OUTPUT_MODIFIER_TOKEN
from cli.utils.constants import SHOW_CONFIG_DNS
from cli.utils.constants import SHOW_CONFIG_GLOBAL
from cli.utils.constants import SHOW_CONFIG_INTERFACE
from cli.utils.constants import DEFAULT_CONFIRMATION_VALUE
from cli.utils.constants import NEWLINE_ASCII_CODE
from cli.utils.error_utils import get_message
from cli.utils.error_utils import render_error
from cli.utils.error_utils import render_error_result
from cli.utils.exceptions import CommandExecutionExeption
from cli.utils.location_utils import FilePaths
from cli.utils.node_utils import is_appliance, is_windows, get_user_id
from cli.utils.render_utils import generate_datetime_stamp_str
from cli.utils.render_utils import get_text_pager
from cli.utils.render_utils import terminal_size
from cli.utils.rest_utils import local_node_request
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import should_enable_central_cli
from compat23.compat23_utils import compat23_string_types
from compat23.compat23_utils import force_unicode


lg = nsx_logging.getLogger("cli.descriptors.cli_command_service")


FORMATTER_REGEX_ARG = "<regex>"
FORMATTER_NUMLINES_ARG = "<number-of-lines>"
FORMATTERS = {
    OUTPUT_MODIFIER_COUNT: {FORMATTER_REGEX_ARG: {"": None}},
    OUTPUT_MODIFIER_FIND: {FORMATTER_REGEX_ARG: {"": None}},
    OUTPUT_MODIFIER_FIRST: {FORMATTER_NUMLINES_ARG: {"": None}},
    OUTPUT_MODIFIER_IGNORE: {FORMATTER_REGEX_ARG: {"": None}},
    OUTPUT_MODIFIER_JSON: {"": None},
    OUTPUT_MODIFIER_LAST: {FORMATTER_NUMLINES_ARG: {"": None}},
    OUTPUT_MODIFIER_MORE: {"": None},
    OUTPUT_MODIFIER_SORT: {"": None},
}

OBFUSCATED_LOG_TOKEN = "<obfuscated>"
# OBFUSCATION_SUFFIX is used to create more informative obfuscated string from
# argument descriptor label. For example, password descriptor will get
# '<password-obfuscated>' as obfuscated string
OBFUSCATION_SUFFIX = OBFUSCATED_LOG_TOKEN[1:]

# TODO: move the following into cli/utils/constants.py
STDIN_INPUT_INVALID_REQUEST_ERROR = 21
STDIN_INPUT_MISSING_INPUT_ERROR = 22
STDIN_INPUT_BAD_ARG_ID_ERROR = 23
STDIN_INPUT_BAD_VERIFY_ARG_ID_ERROR = 24
STDIN_INPUT_BAD_VERIFY_ARG_VALUE_ERROR = 25

_DEFAULT_CENTRAL_CLI_HEADER_WIDTH = 80

_UPGRADE_FLAG = "/image/current_bundle_path"
_UPGRADE_MSG = "/etc/upgrade.msg"

class CommandService(object):
    """Service class that receives/invokes CLI commands and renders output."""

    _NEXT_CONTEXT = {
        "ConfigureTerminal": MODE_CONFIGURE,  # conf t cmd -> configure
        "Interface": MODE_INTERFACE,          # interface cmd -> interface
        "End": MODE_ENABLE                    # end cmd -> enable
    }
    _PREVIOUS_CONTEXT = {
        MODE_CONFIGURE: MODE_ENABLE,          # exit configure -> enable
        MODE_INTERFACE: MODE_CONFIGURE,       # exit interface -> configure
        MODE_VRF: MODE_ENABLE,                # exit vrf -> enable
        MODE_TIER0_SR: MODE_ENABLE,           # exit tier0_sr -> enable
        MODE_TIER1_SR: MODE_ENABLE,           # exit tier1_sr -> enable
        MODE_TIER0_VRF_SR: MODE_ENABLE,       # exit tier0_vrf_sr -> enable
        MODE_PATH: MODE_ENABLE,               # exit path -> enable
        MODE_OFFLINEPATH: MODE_ENABLE,        # exit offlinepath -> enable
        MODE_INTER_SR_VRF: MODE_ENABLE,       # exit inter_sr_vrf -> enable
        MODE_REMOTE_TUNNEL_VRF: MODE_ENABLE   # exit remote_tunnel_vrf -> enable
    }

    def __init__(self, descriptor_factory, prompt=None,
                 current_context=MODE_ENABLE, current_context_data=None):
        self._factory = descriptor_factory
        self._audit_logger = nsx_logging.getLogger("cli.audit")

        # Context and context_data as specified by the last invocation of
        # the invoke_command method.
        self._context = current_context
        self._context_data = current_context_data

        # Modifier string as specified by the last invocation of the
        # invoke_command method or None if no modifier specified.
        self._modifier_str = None

        if prompt:
            self._next_prompt_func = {
                MODE_ENABLE: prompt.enable_prompt,
                MODE_CONFIGURE: prompt.configure_prompt,
                MODE_INTERFACE: prompt.interface_prompt,
                MODE_VRF: prompt.vrf_prompt,
                MODE_TIER0_SR: prompt.tier0_sr_prompt,
                MODE_TIER1_SR: prompt.tier1_sr_prompt,
                MODE_TIER0_VRF_SR: prompt.tier0_vrf_sr_prompt,
                MODE_REMOTE_TUNNEL_VRF: prompt.remote_tunnel_vrf_prompt,
                MODE_PATH: prompt.path_prompt,
                MODE_OFFLINEPATH: prompt.offlinepath_prompt,
                MODE_INTER_SR_VRF: prompt.inter_sr_vrf_prompt
            }
            prompt.service = self
        else:
            self._next_prompt_func = None

        # The _allow_debug_cmds flag is set to True when the user enters the
        # "start debugging" command for the current CLI session.  When True,
        # commands with the debug property become valid commands and support
        # tab completion. Commands with the debug property are never "list"ed.
        self._allow_debug_cmds = False

        # Set service property of all command descriptors to this instance.
        # As Central CLI descriptors are being loaded lazily, we would have to
        # explicitly pass service object, when they are being initialized by
        # CentralCLICacheThread
        for context in self._factory.commands:
            for descriptor in list(self._factory.commands[context].keys()):
                descriptor.service = self

        # Request dictionary populated by invoke_command
        self._req_dict = {}

        self._session_nodes = []  # Set when user enters session mode.
                                  # Cleared when user exits session mode.
        self._session_context_prefix = ''  # Set when user enters session mode
                                           # and invokes a context command.
                                           # Cleared when user exits context
        # whether to show/hide output from navigation commands. Used while
        # getting data from navigation commands to keep a state, within
        # Central CLI session mode.
        self._output_navigate_cmd = False

        # Used while collecting output, to be paged later using text pager. As
        # of now this is only used when OUTPUT_MODIFIER_MORE is used with
        # Central CLI
        self.formatted_output = []
        self._header_width = _DEFAULT_CENTRAL_CLI_HEADER_WIDTH

        # Stores the query arguments as per 'arg_name' for
        # user cmd under execution in case of central cli cmds.
        # Need to clear with every user request.
        # Key: query argument name, Value: remote allowed values
        self._cached_remote_args = dict()

        # Some newly set system/node settings may not be available in current nsx-cli session,
        # this flag specifies whether restart of nsx-cli is required for these settings
        # to be in effect.
        # Flag can not be reset once set to True in current nsx-cli session,
        # will reset with every new nsx-cli session.
        self._requires_restart = False

    @property
    def cli_config(self):
        if is_appliance() or is_windows():
            fpath = FilePaths.CLI_CONFIG_FILE
        else:
            fpath = FilePaths.HOST_CLI_CONFIG_FILE
        try:
            with open(fpath) as fh:
                return simplejson.load(fh)
        except:
            lg.error("Unable to load NSX CLI config file")
        return {}

    @property
    def session_mode_active(self):
        '''Whether CLI is running in Central CLI session mode'''
        return len(self._session_nodes) > 0

    @property
    def session_nodes(self):
        return self._session_nodes

    @session_nodes.setter
    def session_nodes(self, value):
        self._session_nodes = value

    def _clear_remote_arg_cache(self):
        self._cached_remote_args = dict()

    def _log_command(self, cmd_str, descriptor, start_time=None,
                     audit=False, normalized_cmd_list=None,
                     operation_status=None):
        """Log command.

        Args:
            cmd_str: User specified CLI command string.
            descriptor: Descriptor that matches user command string.
            start_time: Timestamp when command started, if specified duration
                is logged.
            audit: True if log as audit log entry.
            normalized_cmd_list: Normalized list of commands (i.e. user
                specified shortcuts resolved to actual keywords/arg values).
            operation_status: Operation status string.
        """
        def _attach_duration(cmd_str, duration):
            if duration:
                cmd_str = "%s (duration: %0.03fs)" % (cmd_str, duration)
            return cmd_str

        duration = (
            (time.time() - start_time)
            if start_time is not None else None)
        # No matching descriptor found, log user command as is
        if not descriptor:
            # Obfuscate the user input because if a typo caused us not to
            # find the descriptor, logging the entire command may leak
            # sensitive information. For example, we don't want to log:
            # "set user admin passwd H3ll0W0r!d" because the user mistakenly
            # used the keyword "passwd" instead of "password". Instead log:
            # "set ... <obfuscated> ...". But if cmd_str consists of a single
            # word, no need to obfuscate.
            parts = shlex.split(cmd_str)
            if len(parts) > 1:
                cmd_str = (
                    "%s ... %s ..." %
                    (shlex.split(cmd_str)[0], OBFUSCATED_LOG_TOKEN))
            cmd_str = _attach_duration(cmd_str, duration)
            if audit:
                if operation_status:
                    cmd_str = "%s, Operation status: %s" % (cmd_str,
                                                            operation_status)
                self._audit_logger.info("CMD: {}".format(cmd_str), audit="true")
            else:
                lg.info("{%s} CMD: %s" %(get_user_id(), cmd_str))
            return
        # Matching descriptor found, scrub user command for sensitive arg values
        # TODO: in crosshairs refactor this so each descriptor/provider can
        # obfuscate their own sensitve arg values because keying off of
        # descriptor.id and command keyword patterns is brittle.
        parts = shlex.split(cmd_str)

        command_indexes = descriptor.get_command_indexes(parts)

        for idx, arg in enumerate(descriptor.path_tuple):
            if is_argument(arg):
                name = resolve_argument_name(arg)
                arg_desc = self._factory.arguments.get(name)
                if arg_desc.obfuscate_in_audit_log:
                    start_idx = sum([x[1] for x in command_indexes[:idx]])
                    size = command_indexes[idx][1]
                    for i in range(start_idx, start_idx+size):
                        parts[i] = (
                            '%s-%s' % (arg_desc.label[:-1], OBFUSCATION_SUFFIX))

        # If targeted for audit log, resolve actual command. We want to show
        # the actual command in addition to the shortcut command in audit log.
        log_actual_command = False
        if audit:
            # Allocate and pre-populate buffer that we'll use to render the
            # actual command based on the cmd_str the user has entered. Note
            # that if cmd_str is a Central CLI command, we want to consolidate
            # all UUIDs and the CENTRAL_CLI_END_TOKEN as a single argument value
            # to match how it's declared in  descriptor.path_tuple
            actual_command_parts = get_cmd_list_path_tuple(parts)
            desc_path = descriptor.path_tuple
            # Ignore substitution for "exit" commands:
            if descriptor.id not in ["Exit", "Exit%s" % CENTRAL_CLI_STR]:
                for pos in range(len(desc_path)):
                    if (not is_argument(desc_path[pos]) and
                            actual_command_parts[pos] != desc_path[pos]):
                        log_actual_command = True
                        actual_command_parts[pos] = desc_path[pos]
                    elif (is_argument(desc_path[pos]) and
                            normalized_cmd_list and
                            actual_command_parts[pos] !=
                            normalized_cmd_list[pos] and
                            OBFUSCATION_SUFFIX not in actual_command_parts[pos]):
                        log_actual_command = True
                        actual_command_parts[pos] = normalized_cmd_list[pos]
        cmd_str = " ".join(parts)
        cmd_str = _attach_duration(cmd_str, duration)
        if log_actual_command:
            cmd_str = (
                "%s (command: %s)" % (cmd_str, " ".join(actual_command_parts)))
        if audit:
            if operation_status:
                cmd_str = "%s, Operation status: %s" % (cmd_str,
                                                        operation_status)
            self._audit_logger.info("CMD: {}".format(cmd_str), audit="true")
        else:
            lg.info("{%s} CMD: %s" %(get_user_id(), cmd_str))


    def invoke_command(self, cmd_str, raw_output, context=MODE_ENABLE):
        """Wrapper around inner_invoke_command.

        The only purpose of this method is to provide a clean-up mechanism, so
        the verify arg cache can always be cleared, after a command invocation.
        """
        try:
            return self.inner_invoke_command(cmd_str, raw_output, context)
        finally:
            # Always clear verify arg cache
            self._factory.verify_arg_cache.clear()
            # clear remote arg cache
            self._clear_remote_arg_cache()


    def inner_invoke_command(self, cmd_str, raw_output, context):
        """Invokes user specified CLI command string.

        Args:
            cmd_str: User specified CLI command string.
            raw_output: True, to print provider output as is (ie. no rendering.)
            context: Current nav context, ie. enable, configure or interface.
        Returns:
            (result, context) tuple where result is CMD_EXECUTED,
            CMD_NOT_EXECUTED, CMD_EXECUTED_REQUESTED_EXIT,
            CMD_EXECUTED_WITH_ERROR_RESULT and context is the
            navigation context after the command executed.
        """
        descriptor = None
        partial_descriptors = []
        # Separate output modifier specifications from actual command. Note
        # we only split on the first occurrence of the output modifier. This
        # is so we support things like "some command | find \|" (i.e. search
        # command output for |.
        cmd_str = force_unicode(cmd_str)
        parts = cmd_str.split(OUTPUT_MODIFIER_TOKEN, 1)
        cmd_str_without_modifier = parts[0]
        modifier_str = parts[1].strip() if len(parts) > 1 else None
        cmd_list = shlex.split(cmd_str_without_modifier)

        if (should_enable_central_cli() and
            is_central_cli_command(cmd_list) and
                (not self._factory.central_cli_descriptors_loaded or
                 not self._factory.node_cache_available)):
            render_error("MSG_CLI_CENTRAL_CLI_STILL_LOADING")
            self._log_command(cmd_str, None)
            self._log_command(cmd_str, None, start_time=time.time(),
                              audit=True,
                              operation_status="CMD_CENTRAL_CLI_STILL_LOADING")
            return CMD_NOT_EXECUTED, context

        # Don't verify arguments (due to latency) for Central CLI commands
        verify_argument = (False if cmd_list[0] == CENTRAL_CLI_START_TOKEN
                           else True)
        self._req_dict = {"cmd_list": cmd_list, "context": context}
        if cmd_list and cmd_list[0] == "#":
            self._log_command(cmd_str, None)
            self._log_command(cmd_str, None, start_time=time.time(),
                              audit=True, operation_status="CMD_EXECUTED")
            return CMD_EXECUTED, context
        else:
            exact_matches, partial_matches, incomp_matches, value_mismatches = (
                self.resolve_command_descriptor(
                    cmd_list, context, allow_unauthorized=True,
                    verify_argument=verify_argument))
            # Note, *_matches can contain command descriptors that are
            # unauthorized, so the following code that tries to derive
            # which command descriptor the user wants to invoke (accounting
            # for partial and incomplete matches), needs to account for and
            # ignore these as necessary. Unauthorized commands are not
            # filtered out because if matched command is not authorized,
            # we need to output an unauthorized error.
            # If more than one exact match found, ignore hidden commands
            if len(exact_matches) > 1:
                # If more than one exact match found, ignore hidden and
                # unauthorized commands
                exact_matches = [
                    d for d in exact_matches if not d.hidden and d.authorized]
            if len(exact_matches) > 1 and not verify_argument:
                # As verify_argument was disabled for Central CLI, and we got
                # multiple exact_matches, its time to enable verify_argument,
                # but only for matching descriptors
                exact_matches = [
                    d for d in exact_matches if
                    d.match_command_list(cmd_list, True) == CMD_MATCH_EXACT]

            if len(exact_matches) == 1:
                descriptor = exact_matches[0]
            elif len(exact_matches) == 0 and len(partial_matches) > 1:
                for desc in partial_matches:
                    if (desc.hidden or desc.debug or not desc.authorized or
                            not self._allow_command_descriptor(desc, context)):
                        continue
                    partial_descriptors.append(desc)
                if len(partial_descriptors) == 1:
                    partial_matches = partial_descriptors
                    partial_descriptors = []
            if len(exact_matches) == 0 and len(partial_matches) == 1:
                descriptor = partial_matches[0]
            elif (len(exact_matches) == 0 and len(partial_matches) == 0 and
                  len(value_mismatches) > 0):
                # Check if any query or verify args returned an error
                pos = 0
                for descriptor in value_mismatches:
                    for item in descriptor.path_tuple:
                        if pos >= len(cmd_list):
                            break
                        user_input = cmd_list[pos]
                        pos += 1
                        if not is_argument(item):
                            continue
                        arg_desc = self._get_argument_descriptor(item)
                        if arg_desc.verify_argument:
                            result = arg_desc.match_verify_arg_value(
                                user_input,
                                self._req_dict,
                                false_if_error=False,
                                debug_mode=self._allow_debug_cmds,
                                remote_nodes=self.session_nodes\
                                    if self.session_mode_active else None)
                        elif arg_desc.query_argument:
                            result = arg_desc.get_allowed_values(
                                self._req_dict,
                                empty_list_if_error=False,
                                debug_mode=self._allow_debug_cmds,
                                session_nodes=self.session_nodes)
                        else:
                            continue
                        # If result is a string, it means that a verify or
                        # query argument wants to show a custom error
                        if isinstance(result, str):
                            render_error_result(result)
                            self._log_command(cmd_str, None)
                            self._log_command(
                                cmd_str, None, start_time=time.time(),
                                audit=True, operation_status="CMD_NOT_EXECUTED")
                            return CMD_NOT_EXECUTED, context
                descriptor = value_mismatches[0]
                invalid_values = descriptor.invalid_argument_values(
                    tuple(cmd_list))
                if invalid_values:
                    arg_name = invalid_values[0][0]
                    arg_desc = self._get_argument_descriptor(arg_name)
                    arg_value = invalid_values[0][1]
                    # Invalid argument value
                    render_error(
                        "MSG_CLI_INVALID_ARGUMENT", arg_desc.label, arg_value)
                    show_argument_description(arg_desc)
                    # TODO: support raw_output
                    self._log_command(cmd_str, None)
                    self._log_command(
                        cmd_str, None, start_time=time.time(),
                        audit=True, operation_status="CMD_INVALID_ARGUMENT")
                    return CMD_NOT_EXECUTED, context
                else:
                    descriptor = None

        if (len(exact_matches) > 1 and is_central_cli_command(cmd_list)):
            # As Central CLI command execution relies on path_tuple only, any
            # descriptors with matching path_tuple can be assumed to be same,
            # and we can ignore all but one. However, each matching descriptor
            # should have no common node_types, as that is an erroneous
            # condition
            node_types = [m.node_types for m in exact_matches]
            if all([set(l[0]).isdisjoint(l[1])
                    for l in permutations(node_types, 2)]):
                descriptor = exact_matches[0]

        if not descriptor and partial_descriptors:
            self._log_command(cmd_str, None)
            self._log_command(cmd_str, None, start_time=time.time(),
                              audit=True, operation_status="CMD_AMBIGUOUS")
            return self.ambiguous_command_found(
                cmd_str, None, raw_output, context, True)
        # Check if user not authorized to invoke command
        elif descriptor and not descriptor.authorized:
            self._log_command(cmd_str, None)
            lg.info(
                "User '%s' attempted to invoke an unauthorized command",
                getpass.getuser())
            self._log_command(cmd_str, None, start_time=time.time(),
                              audit=True, operation_status="CMD_UNAUTHORIZED")
            render_error("MSG_CLI_UNAUTHORIZED")
            return CMD_NOT_EXECUTED, context
        elif (descriptor and
              descriptor.requires_tty_if_remote and
              (is_central_cli_command(cmd_list) or self.session_mode_active) and
              descriptor.id != CENTRAL_CLI_STR):
            self._log_command(cmd_str, None)
            render_error("MSG_CLI_CMD_DOES_NOT_SUPPORT_CENTRAL_CLI")
            return CMD_NOT_EXECUTED, context
        # Check if incomp_mismatches exist and if
        elif (not descriptor or
                not self._allow_command_descriptor(descriptor, context)):
            self._log_command(cmd_str, None)
            self._log_command(cmd_str, None, start_time=time.time(),
                              audit=True, operation_status="CMD_NOT_FOUND")
            return self.command_not_found(cmd_str, raw_output, context)
        # If command has enum/query args, make sure user values for these args
        # aren't ambiguous (e.g. "protocol t" where protocol can be tcp or tls)
        pos = 0
        for item in descriptor.path_tuple:
            user_input = cmd_list[pos]
            pos += 1
            if is_argument(item):
                arg_desc = self._get_argument_descriptor(item)
                if (arg_desc.enum_argument or arg_desc.query_argument):
                    values_found = []
                    # Get the value from cache if available, before getting values from remote node.
                    if arg_desc.query_argument and arg_desc.name in self._cached_remote_args:
                        values = self._cached_remote_args[arg_desc.name]
                    else:
                        values = self._get_argument_values(arg_desc.name)
                    for value in values:
                        # If there's an exact match, then it's not ambiguous
                        if value == user_input:
                            values_found = []
                            break
                        elif value.startswith(user_input):
                            values_found.append(value)
                    if len(values_found) > 1:
                        return self.ambiguous_command_value_found(
                            arg_desc, user_input, values_found, context)
        # Display upgrade warning if upgrade in progress
        if (os.path.isfile(_UPGRADE_FLAG) and cmd_str != "exit" and
            not cmd_str.startswith("resume upgrade-bundle")):
           with open(_UPGRADE_MSG) as fd:
               um = fd.read()
               if um:
                   sys.stdout.write(um)
                   sys.stdout.write("\n")
        # Invoke the command
        self._log_command(cmd_str, descriptor, audit=False)
        start_time = time.time()
        if modifier_str is not None and len(modifier_str) == 0:
            render_error("MSG_CLI_OUTPUT_MODIFIER_UNSPECIFIED")
            self._log_command(
                cmd_str, descriptor, start_time=time.time(),
                audit=True, operation_status="CMD_OUTPUT_MODIFIER_UNSPECIFIED")
            return CMD_NOT_EXECUTED, context
        is_nav, new_context = self._navigate_command(
            descriptor, context, cmd_list)
        if is_nav:
            if (is_central_cli_command(cmd_list) and
                    not self.session_mode_active):
                render_error("MSG_CLI_CMD_REQUIRES_SESSION_MODE")
                return CMD_NOT_EXECUTED, context
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, operation_status="CMD_EXECUTED_REQUESTED_EXIT")
            if new_context != MODE_QUIT:
                return CMD_EXECUTED, new_context
            return CMD_EXECUTED_REQUESTED_EXIT, MODE_QUIT

        # Return error if user specified a modifier
        # but command has suppress_modifiers: true
        if descriptor.suppress_modifiers and modifier_str:
            render_error("MSG_CLI_OUTPUT_MODIFIERS_NOT_ALLOWED")
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, operation_status="CMD_OUTPUT_MODIFIERS_NOT_ALLOWED")
            return CMD_NOT_EXECUTED, context
        # Validate modifier_str
        valid_modifier, normalized_modifier_str = self._validate_modifier_str(
            modifier_str)
        if not valid_modifier:
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, operation_status="CMD_NOT_VALID_MODIFIER")
            return CMD_NOT_EXECUTED, context

        self._modifier_str = normalized_modifier_str

        normalized_cmd_list = []
        try:
            self._set_command_start_time(start_time)
            result = self._invoke_command(
                descriptor, cmd_list, raw_output, normalized_modifier_str,
                context, normalized_cmd_list, verify_argument)
        except CommandExecutionExeption as ex:
            parts = str(ex).strip().split("\n")
            msg_id = parts[0].strip()
            lg.warn(get_message(msg_id, *parts[1:]))
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, normalized_cmd_list=normalized_cmd_list,
                operation_status="CMD_EXECUTED_WITH_ERROR_RESULT")
            return CMD_EXECUTED_WITH_ERROR_RESULT, context
        except CentralCLIAPIException:
            render_error("MSG_CLI_UNEXPECTED_ERROR_PROTON")
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, normalized_cmd_list=normalized_cmd_list,
                operation_status="CMD_UNEXPECTED_ERROR_PROTON")
            return CMD_NOT_EXECUTED, context
        except EOFError:
            render_error("MSG_CLI_UNEXPECTED_ERROR")
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, normalized_cmd_list=normalized_cmd_list,
                operation_status="CMD_UNEXPECTED_ERROR")
            return CMD_NOT_EXECUTED, context
        else:
            self._log_command(
                cmd_str, descriptor, start_time=start_time,
                audit=True, normalized_cmd_list=normalized_cmd_list,
                operation_status="CMD_EXECUTED")
        finally:
            self._reset_command_start_time()
        return result

    def _set_command_start_time(self, start_time):
        self._command_start_time = start_time

    def _reset_command_start_time(self):
        self._command_start_time = None

    def get_command_start_time(self):
        return self._command_start_time

    def command_not_found(self, cmd_str, raw_output=None, context=MODE_ENABLE):
        """Process command not found error.

        Args:
            cmd_str: User specified CLI command string.
            raw_output: True, to print provider output as is (ie. no rendering.)
            context: Current nav context, ie. enable, configure or interface.
        Returns:
            (result, context) tuple where result is CMD_NOT_EXECUTED,
            MSG_CLI_COMMAND_NOT_FOUND and context is the navigation context
            after the command executed.
        """
        if raw_output:
            return "MSG_CLI_COMMAND_NOT_FOUND", context
        # Fix for #1469308: display more helpful error message by attempting to
        # match what the user has already entered to all known commands. We
        # match based on the second, and optionally first, user entered keyword
        # ("verb object" per the Transformers CLI Strategy guide) in the
        # following order:
        # 1. List all commands that match first and second keyword.
        # 2. List all commands that match first keyword and first letter
        #    of second keyword.
        # 3. List all commands that match first keyword.
        # If a lower number rule returns matches, don't apply remaining rules.
        # Also, by match, we mean starts with match.
        matches = None
        is_central_cli = False
        cmds = self._get_all_commands(
            context, cmd_str=shlex.split(cmd_str),
            wait_for_central_cli_node_cache=False)
        parts = get_central_cli_user_input(shlex.split(cmd_str))
        if parts:
            is_central_cli = True
        else:
            parts = shlex.split(cmd_str)
        if len(parts) > 1:
            matches_words = []
            matches_letter = []
            key1, key2 = parts[0], parts[1]
            key2_first_letter = key2[0]
            for cmd in cmds:
                cmd_parts = cmd.split()
                if ((is_central_cli or self.session_mode_active)
                                and is_central_cli_command(cmd_parts)):
                    # If current command to check is a central cli command,
                    # re-position cmd_parts so "on <uuid> exec" is ignored
                    cmd_parts = cmd_parts[3:]
                if len(cmd_parts) < 2:
                    continue
                # Check current command for case 1
                elif (cmd_parts[0].startswith(key1) and
                        cmd_parts[1].startswith(key2)):
                        if self.session_mode_active:
                            matches_words.append(" ".join(cmd_parts))
                        else:
                            matches_words.append(cmd)
                # Check current command for case 2
                elif (not matches_words and cmd_parts[0].startswith(key1)
                    and cmd_parts[1].startswith(key2_first_letter)):
                    if self.session_mode_active:
                        matches_letter.append(" ".join(cmd_parts))
                    else:
                        matches_letter.append(cmd)
            matches = matches_words if matches_words else matches_letter
        # If no matches so far, check for case 3
        if not matches and len(parts) > 0:
            search_str = parts[0]
            if is_central_cli or self.session_mode_active:
                for cmd in cmds:
                    cmd_parts = cmd.split()
                    if is_central_cli_command(cmd_parts):
                        # If current command to check is a central cli command,
                        # re-position cmd_parts so "on <uuid> exec" is ignored
                        cmd_parts = cmd_parts[3:]
                        if " ".join(cmd_parts).startswith(search_str):
                            if matches is None:
                                matches = []
                            if is_central_cli:
                                matches.append(cmd)
                            else:
                                matches.append(" ".join(cmd_parts))
            else:
                matches = [cmd for cmd in cmds if cmd.startswith(search_str)]
        if matches:
            matches.sort()
            render_error(
                "MSG_CLI_COMMAND_NOT_FOUND_ALTERNATIVES", cmd_str,
                "\n    ".join(matches))
        else:
            render_error("MSG_CLI_COMMAND_NOT_FOUND", cmd_str)
        return CMD_NOT_EXECUTED, context

    def ambiguous_command_found(self, cmd_str, descriptors=None,
                                raw_output=None, context=MODE_ENABLE,
                                suppress_hidden_debug=False,
                                search_hint=None, custom_error_id=None,
                                custom_alt_error_id=None):
        """Process ambiguous command (i.e. specified command matches multiple
           descriptors) error.

        Args:
            cmd_str: User specified CLI command string.
            descriptors: Descriptors that match the ambiguous command.
            raw_output: True, to print provider output as is (ie. no rendering.)
            context: Current nav context, ie. enable, configure or interface.
            suppress_hidden_debug: Set True to not show hidden or debug
                descriptors.
            search_hint: If descriptors not specified, search for ambiguous
                descriptors based on this hint instead of cmd_str.
            custom_error_id: Custom error ID to use when no alternatives found.
            custom_alt_error_id: Custom error ID to use when alternative
                commands found.
        Returns:
            (result, context) tuple where result is CMD_NOT_EXECUTED,
            MSG_CLI_COMMAND_NOT_FOUND and context is the navigation context
            after the command executed.
        """
        if raw_output:
            return "MSG_CLI_COMMAND_NOT_FOUND", context
        # If descriptors not specified, search for all descriptors that match
        # search_hint or cmd_str if search_hint not specified
        if not descriptors:
            hint = search_hint if search_hint else cmd_str
            cmd_list = hint.split(OUTPUT_MODIFIER_TOKEN, 1)[0].split()
            exact_matches, partial_matches, incomplete_matches, _ = (
                self.resolve_command_descriptor(cmd_list, context))
            descriptors = exact_matches + partial_matches + incomplete_matches
        cmd_list = shlex.split(cmd_str)
        central_cli_nodes = []
        if is_central_cli_command(cmd_list):
            central_cli_nodes = get_central_cli_nodes(cmd_list)
        commands = []
        for descriptor in descriptors:
            if (suppress_hidden_debug and
                    (descriptor.hidden or descriptor.debug or
                     descriptor.doc_hidden or
                    (descriptor.id.endswith(CENTRAL_CLI_STR) and
                            descriptor.requires_tty_if_remote))):
                continue
            # Check descriptor applicable on central cli nodes, if specified
            if (central_cli_nodes and
                    not descriptor.allowed_on_nodes(central_cli_nodes)):
                continue
            # Add current descriptor to list of suggestions
            if descriptor.doc_override:
                if self.session_mode_active:
                    cmd = " ".join(descriptor.doc_override.split()[3:])
                    commands.append(cmd)
                else:
                    commands.append(descriptor.doc_override)
            else:
                cmd = []
                for item in descriptor.path_tuple:
                    if is_argument(item):
                        arg_desc = self._get_argument_descriptor(item)
                        item = arg_desc.label
                        # Append CENTRAL_CLI_END_TOKEN for Central CLI commands
                        if arg_desc.id == 'RegisteredNodeUuid':
                            item = '%s %s' % (item, CENTRAL_CLI_END_TOKEN)
                    cmd.append(item)
                commands.append(" ".join(cmd))
        if commands:
            commands.sort()
            err_id = (
                custom_alt_error_id if custom_alt_error_id else
                "MSG_CLI_AMBIGUOUS_COMMAND_ALTERNATIVES")
            render_error(err_id, cmd_str, "\n    ".join(commands))
        else:
            err_id = (
                custom_error_id if custom_error_id else
                "MSG_CLI_AMBIGUOUS_COMMAND")
            render_error(err_id, cmd_str)
        return CMD_NOT_EXECUTED, context

    def ambiguous_command_value_found(self, arg_desc, user_input,
                                      possible_values, context):
        """Process ambiguous command value error (i.e. specified command has
           argument value that satisfies more than one allowed enum or query
           argument values).

        Args:
            arg_desc: Argument descriptor.
            user_input: User input that triggered ambiguous value error.
            possible_values: Possible argument values.
            context: Current nav context, ie. enable, configure or interface.
        Returns:
            (CMD_NOT_EXECUTED, context) where context is the navigatio context
            after the command is exectued.
        """
        max_len = 8
        options = {}
        for value in possible_values:
            desc = arg_desc.get_argument_value_desc(value, self._req_dict)
            options[value] = desc
            max_len = len(value) if len(value) > max_len else max_len
        sorted_value_desc = []
        sorted_values = list(options.keys())
        sorted_values.sort()
        for value in sorted_values:
            fmt = "%%-%ds  %%-s" % max_len
            sorted_value_desc.append(fmt % (value, options[value]))
        render_error(
            "MSG_CLI_AMBIGUOUS_COMMAND_VALUE_ALTERNATIVES", user_input,
            "\n    ".join(sorted_value_desc))
        return CMD_NOT_EXECUTED, context

    def resolve_command_descriptor(self, cmd_list, context,
                                   allow_unauthorized=False,
                                   verify_argument=True,
                                   input_has_trailing_space=False):
        """Return descriptors that match exactly, partially or mismatched args.

        Args:
            cmd_list: List of cmd tokens to match (e.g. ["show", "version"])
            context: Current nav context, ie. enable, configure or interface.
            allow_unauthorized: Set True to ignore descriptor's authorized
                property.
            input_has_trailing_space: Flag specifies user input ends with
                trailing space(s).
        Return:
            (exact_matches, partial_matches, incomplete_matches,
            value_mismatches) tuple where:
            - exact_matches contain all command descriptors where the
              cmd_list exactly matches a descriptors path property.
            - partial_matches contain all command descriptors where cmd_list
              partially match a descriptors path when accounting for command
              shortcuts (e.g. "sh ver" partially matches "show version").
            - incomplete matches contain all command descriptors where cmd_list
              had too few keywords but what keywords were specified, matched
              exactly.
            - value_mismatches contain all command descriptors where cmd_list
              mismatched due to an argument (e.g. "set mtu abc" value
              mismatches a descriptor with path: "set mtu <integer>").
        Notes:
            - The return lists will only contain a descriptor if that
              descriptor is available in the specified context.
        """
        cmd_list_tuple = tuple(cmd_list)
        exact_matches = []
        partial_matches = []
        incomplete_matches = []
        value_mismatches = []
        if context not in self._factory.commands:
            return (exact_matches, partial_matches, incomplete_matches,
                    value_mismatches)

        # Match specified cmd_list with list of known commands
        for descriptor in list(self._factory.commands[context].keys()):
            # If the descriptor's context_available setting does not
            # include the specified context, skip matching this descriptor
            if not self._allow_command_descriptor(
                    descriptor, context, allow_unauthorized):
                continue
            # Note that we allow hidden commands to be matched, rejecting
            # hidden commands here means hidden commands cannot be executed
            result = descriptor.match_command_list(cmd_list_tuple,
                                                   verify_argument,
                                                   input_has_trailing_space)
            if result == CMD_MATCH_EXACT:
                exact_matches.append(descriptor)
            elif result == CMD_MATCH_PARTIAL:
                partial_matches.append(descriptor)
            elif result == CMD_MATCH_INCOMPLETE:
                incomplete_matches.append(descriptor)
            elif result == CMD_MATCH_VALUE_MISMATCH:
                value_mismatches.append(descriptor)
        return (exact_matches, partial_matches, incomplete_matches,
                value_mismatches)

    def _allow_command_descriptor(self, descriptor, context,
                                  allow_unauthorized=False):
        """Return True if specified command descriptor is allowed.

        Allowed means the descriptor must not have the debug property set
        to True when self._allow_debug_cmds is False, and the descriptor's
        context_available property must include the current context

        Args:
            descriptor: CLI command descriptor to test whether allowed.
            context: Current context to test whether descriptor allowed.
            allow_unauthorized: Set True to ignore descriptor's authorized
                property.
        Returns:
            True, if specified allowed; otherwise, False.
        """
        if descriptor.debug and not self._allow_debug_cmds:
            return False
        elif not allow_unauthorized and not descriptor.authorized:
            return False
        elif MODE_ALL in descriptor.context_available:
            return True
        elif context in descriptor.context_available:
            return True
        return False

    def _validate_modifier_str(self, modifier_str):
        """Validate user entered output modifier string.

        Args:
            modifier_str: Output modifier string to validate.
        Returns:
            (True, normalized_modifier_str) if user specified output modifier
            is valid; otherwise, (False, None)
        """
        if not modifier_str:
            return True, modifier_str
        parts = modifier_str.split()
        # Auto-complete modifier if it's partially specified
        possible_modifiers = []
        for formatter in FORMATTERS:
            if formatter.startswith(parts[0]):
                possible_modifiers.append(formatter)
        len_possible_modifiers = len(possible_modifiers)
        if len_possible_modifiers == 1:
            modifier_str = modifier_str.replace(
                parts[0], possible_modifiers[0], 1)
            parts[0] = possible_modifiers[0]
        elif len_possible_modifiers > 1:
            if len_possible_modifiers == 2:
                options = " or ".join(possible_modifiers)
            else:
                options = (
                    ", ".join(possible_modifiers[:-1]) + " or " +
                    possible_modifiers[-1])
            render_error(
                "MSG_CLI_OUTPUT_MODIFIER_MULTIPLE_MATCHED", parts[0], options)
            return False, None
        # Check valid output modifier
        if parts[0] not in FORMATTERS:
            render_error("MSG_CLI_OUTPUT_MODIFIER_INVALID", parts[0])
            return False, None
        # Check arguments haven't been specified for formmatters
        # that don't accept arguments
        elif (parts[0] in [OUTPUT_MODIFIER_JSON, OUTPUT_MODIFIER_MORE,
                           OUTPUT_MODIFIER_SORT] and len(parts) > 1):
            render_error("MSG_CLI_OUTPUT_MODIFIER_NO_ARGS", parts[0])
            return False, None
        # Check only one argument specified for output modifier that take them
        elif len(parts) > 2:
            render_error("MSG_CLI_OUTPUT_MODIFIER_TOO_MANY_ARGS",
                         parts[0], " ".join(parts[1:]))
            return False, None
        # Check argument specified for output modifier requiring regex
        if (parts[0] in [OUTPUT_MODIFIER_COUNT, OUTPUT_MODIFIER_FIND,
                         OUTPUT_MODIFIER_IGNORE] and len(parts) < 2):
            render_error("MSG_CLI_OUTPUT_MODIFIER_MISSING_ARG", parts[0])
            arg_desc = self._get_argument_descriptor(FORMATTER_REGEX_ARG)
            if arg_desc:
                show_argument_description(arg_desc)
            return False, None
        # Check argument specified for output modifier requiring regex
        if parts[0] in [OUTPUT_MODIFIER_FIRST, OUTPUT_MODIFIER_LAST]:
            if len(parts) < 2:
                render_error("MSG_CLI_OUTPUT_MODIFIER_MISSING_ARG", parts[0])
                arg_desc = self._get_argument_descriptor(FORMATTER_NUMLINES_ARG)
                if arg_desc:
                    show_argument_description(arg_desc)
                return False, None
            else:
                try:
                    val = int(parts[1])
                    if val < 0:
                        raise ValueError()
                except ValueError:
                    render_error(
                        "MSG_CLI_OUTPUT_MODIFIER_INVALID_ARG", parts[1])
                    arg_desc = self._get_argument_descriptor(
                        FORMATTER_NUMLINES_ARG)
                    if arg_desc:
                        show_argument_description(arg_desc)
                    return False, None
        return True, modifier_str

    def render_central_cli_data(self, data):
        """Callback function to render CentralCLI data from each remote node"""

        def _format_header(uuid):
            node_name = self._factory.node_cache[uuid]['display_name']
            type_label = self._factory.node_cache[uuid].get(
                "node_type_label", "")
            parts = [
                "-" * self._header_width,
                "  ".join([uuid, type_label, node_name]),
                "-" * self._header_width]
            return "\n".join(parts)

        uuid = data[1].lower()  # Node cache uses lowercase UUIDs as keys
        modifier = data[3]
        start_time = data[4]
        if modifier == OUTPUT_MODIFIER_JSON:
            out = re.sub(r'\n', '\n  ', data[2])
            out = re.sub(r'^', '  ', out)
            sys.stdout.write('\n"%s":\n' % uuid)
            sys.stdout.write('%s\n' % out.rstrip())
        elif modifier == OUTPUT_MODIFIER_MORE:
            self.formatted_output.append('%s\n' % _format_header(uuid))
            # Print start_time of executing command if available in output
            if start_time is not None and\
               self.cli_config.get('cli_output_datetime', False):
                self.formatted_output.append(
                    '%s\n' % generate_datetime_stamp_str(start_time))
            self.formatted_output.append('%s\n' % data[2])
        else:
            # TODO This code should use render_utils
            sys.stdout.write('%s\n' % _format_header(uuid))

            # Print start_time of executing command if available in output
            if start_time is not None and\
               self.cli_config.get('cli_output_datetime', False):
                sys.stdout.flush()
                sys.stderr.write(
                    '%s\n' % generate_datetime_stamp_str(start_time))
                sys.stderr.flush()
            sys.stdout.write('%s\n' % data[2])

    def _invoke_command(self, descriptor, cmd_list, raw_output, modifier_str,
                        context, normalized_cmd_list, verify_argument):
        """Invokes user specified CLI command string using specified descriptor.

        Args:
            descriptor: CLI command descriptor to invoke.
            cmd_list: User specified CLI command list.
            raw_output: True, to print provider output as is (ie. no rendering.)
            modifier_str: Output modifier string to apply.
            context: Current nav context, ie. enable, configure or interface.
            normalized_cmd_list: List buffer where normalized command list is
                copied to.
        Returns:
            (result, context) tuple where result is CMD_EXECUTED,
            CMD_NOT_EXECUTED, CMD_EXECUTED_REQUESTED_EXIT,
            CMD_EXECUTED_WITH_ERROR_RESULT and context is the
            navigation context after the command executed.
        """
        cmd_list = descriptor.normalize_command_list(cmd_list, verify_argument)
        normalized_cmd_list[:] = []  # Clear existing contents of list
        for item in cmd_list:
            normalized_cmd_list.append(item)
        cmd_list = self._query_interactive_args(descriptor, cmd_list)
        # TODO: Find a better way to identify central CLI command. While loading
        # descriptors in descriptor_factory, all the command descriptors, which
        # originally didn't belong to the node, were loaded with 'CentralCLI'
        # added to their descriptor id
        if descriptor.id == CENTRAL_CLI_STR:  # entering session mode
            # As nodes will be a single string, split it
            nodes = get_central_cli_nodes([cmd_list[0]] + cmd_list[1].split() +
                                          cmd_list[2:])

            # For DropKick, only allow 1 node in session mode
            if len(nodes) > 1:
                render_error("MSG_CLI_CMD_DOES_NOT_SUPPORT_MULTIPLE_NODES")
                return CMD_NOT_EXECUTED, context

            sys.stdout.write('Entering session mode\n')
            nodes = [self._factory.node_cache[node]['uuid'] for node in nodes]
            self.session_nodes = nodes

            return self._render_output("OK\n\n%s" % "DONE", descriptor,
                                       cmd_list, True, modifier_str, context)
        elif descriptor.id.endswith(CENTRAL_CLI_STR):
            # example cmd_list: ['on', 'host1 host2 edge1
            #                       CENTRAL_CLI_END_TOKEN', 'get', 'hostname']
            if modifier_str == 'json':
                sys.stdout.write(('{\n'))
            if not self.session_mode_active:
                nodes = cmd_list[1].split()[:-1]
                nodes = [(
                    self._factory.node_cache[node]['uuid'],
                    self._factory.node_cache[node]['node_type'])
                    for node in nodes]
                cmd = ' '.join(cmd_list[2:])
            else:  # In session mode
                # May contain capitalized uuid for managers
                nodes = self.session_nodes
                nodes = [(
                    v['uuid'], v['node_type'])
                    for v in list(self._factory.node_cache.values())
                    if v['uuid'] in nodes]
                cmd = ' '.join(cmd_list)
            # When invoking a remote command, check if session mode is active
            # and whether a context is active within the session. If so, the
            # remote invocation has to first enter the context before invoking
            # the command specified by the user (because remote invocations
            # are done via the equivalent of: nsxcli -c <user_cmd>)
            if self._session_context_prefix:
                cmd = self._session_context_prefix + ';' + cmd
            remote_context = MODE_ENABLE  # Because we invoke using nsxcli -c
                                          # remote context starts at "enable"
            # Calculate header width
            cols = terminal_size()[1]
            self._header_width = (
                cols if cols else _DEFAULT_CENTRAL_CLI_HEADER_WIDTH)
            async_invoke_and_wait(nodes, cmd, modifier_str, remote_context,
                                  getpass.getuser(),
                                  [self._factory._user_permissions],
                                  self._allow_debug_cmds,
                                  self.render_central_cli_data)
            if modifier_str == 'json':
                sys.stdout.write(('\n}\n'))
            if modifier_str == OUTPUT_MODIFIER_MORE:
                pager = get_text_pager()
                if pager:
                    try:
                        pager.write(''.join(self.formatted_output))
                    except (IOError, KeyboardInterrupt):
                        pass
                    finally:
                        pager.close()
            self.formatted_output = []  # reset formatted_output
            return self._render_output("OK\n\n%s" % "DONE", descriptor,
                                       cmd_list, True, modifier_str, context)
        else:
            output = descriptor.invoke(cmd_list)
            return self._render_output(output, descriptor, cmd_list, raw_output,
                                       modifier_str, context)

    def _render_output(self, output, descriptor, cmd_list, raw_output,
                       modifier_str, context):
        """Print command output as returned by command descriptor provider.

        Args:
            output: Output returned by descriptor provider.
            descriptor: CLI command descriptor to invoke.
            cmd_list: User specified CLI command list.
            raw_output: True, to print provider output as is (ie. no rendering.)
            modifier_str: Output modifier string to apply.
            context: Current nav context, ie. enable, configure or interface.
        Returns:
            (result, context) tuple where result is CMD_EXECUTED,
            CMD_NOT_EXECUTED, CMD_EXECUTED_REQUESTED_EXIT,
            CMD_EXECUTED_WITH_ERROR_RESULT and context is the
            navigation context after the command executed.
        """
        if output is None:
            return CMD_EXECUTED, context
        # If raw_output specified, return provider output as is
        if raw_output:
            return output, context
        # All CLI command providers must return a string output in the form of:
        # header\n
        # \n
        # result
        #
        # The header portion is in the form of:
        # return_code\n
        # format_arg: some string\n
        # format_arg: some other string\n
        #
        # If return_code is not "0", it is used as a key to determine an
        # error string pattern which can contain zero or more arguments (%s)
        # which the framework replaces with the specified format_arg(s).
        # Note: format_arg is the only "header" currently supported, future
        # versions may support additional headers.
        #
        parts = output.split("\n\n", 1)
        # The return code is the first line of the header section
        return_code = parts[0].split("\n")[0].strip()
        if return_code == "MSG_CLI_EXITING":
            return CMD_EXECUTED_REQUESTED_EXIT, context
        if return_code != "OK":
            output_prefix = None
            if self.cli_config.get('cli_output_datetime', False) and\
               not (modifier_str and\
                    modifier_str.startswith(OUTPUT_MODIFIER_JSON)):
                # If node properties have changed, then
                # get the system datetime in case if restart of nsx-cli session is required.
                start_time = None
                if not self.get_requires_restart():
                    start_time = self.get_command_start_time()
                output_prefix = generate_datetime_stamp_str(start_time)
            render_error_result(parts[0], output_prefix=output_prefix)
            raise CommandExecutionExeption(parts[0])
        # The result portion is a json formatted string
        if len(parts) != 2:
            msg = ("Invalid output from provider, cmd: %s   output: %s" %
                   (cmd_list, output))
            assert False, msg
            lg.error(msg, error_code=ec.OPERATION_FAILED)
            return CMD_EXECUTED, context
        # Provider returned a "0" return_code, render results
        try:
            descriptor.render(parts[1], modifier_str)
        except re.error as ex:
            lg.error("Invalid find expression specified: %s", modifier_str,
                     error_code=ec.INVALID_ARGUMENT)
            render_error("MSG_CLI_OUTPUT_MODIFIER_INVALID_ARG", modifier_str)
        return CMD_EXECUTED, context

    def _query_interactive_args(self, descriptor, cmd_list):
        """Query user for arg values if descriptor specifies interactive args.

        Args:
            descriptor: CLI command descriptor to invoke.
            cmd_list: User specified CLI command list.
        Returns:
            cmd_list where interactive arg values (if any) are appended at
            the end of the command list.
        """
        arg_names = list(descriptor.interactive_args.keys())
        arg_names.sort()
        for arg_name in arg_names:
            arg = descriptor.interactive_args[arg_name]
            while True:
                if arg.password_arg:
                    value = getpass.getpass(arg.arg_request_string)
                else:
                    sys.stderr.write(arg.arg_request_string)
                    sys.stderr.flush()
                    value = sys.stdin.readline()
                    if value is None:
                        continue
                    elif value[-1] == "\n":

                        # Check for newline character when user does
                        # not provide confirmation and default to no.
                        if len(value) == 1 and ord(value) == NEWLINE_ASCII_CODE:
                            value = DEFAULT_CONFIRMATION_VALUE
                        else:
                            value = value[:-1]
                if not arg.arg_pattern or re.match(arg.arg_pattern, value):
                    arg.value = value
                    cmd_list.append(value)
                    break
                elif arg.invalid_arg_msg:
                    sys.stderr.write(arg.invalid_arg_msg)
                    sys.stderr.write("\n")
                    sys.stderr.flush()
                    sys.stderr.write("\n")
        return cmd_list

    def _get_all_commands(self, context=MODE_ENABLE, cmd_str=[],
                          wait_for_central_cli_node_cache=True):
        """Return list of supported commands (in space delimited format)

        Args:
            context: Current context.
        Note:
            - Only commands whose descriptor.context_available properties
              contain the specified context will be returned.
            - Command arguments are represented by argument labels.
              e.g. "ping ip IPADDRESS" instead of "ping ip <ipaddress>"
        """
        result = set()
        if context not in self._factory.commands:
            return []

        items = list(self._factory.commands[context].items())
        # If user specified a Central CLI command, determine UUIDs of specified
        # nodes so their node types can be used to filter out commands below.
        node_types = []
        central_cli_nodes = []
        if is_central_cli_command(cmd_str):
            central_cli_nodes = get_central_cli_nodes(cmd_str)
        if self.session_mode_active:
            central_cli_nodes = self._session_nodes
        for descriptor, path_tuple in items:
            if (descriptor.hidden or descriptor.doc_hidden or descriptor.debug
                    or not self._allow_command_descriptor(descriptor, context)
                    or not descriptor.authorized
                    or (descriptor.id.endswith(CENTRAL_CLI_STR)
                    and descriptor.requires_tty_if_remote)):
                continue
            # Hide context command if running as central CLI command
            if (is_central_cli_command(cmd_str)
                    and descriptor.id.endswith(CENTRAL_CLI_STR)
                    and descriptor.context_command):
                continue
            # Check descriptor applicable on central cli nodes, if specified
            if (central_cli_nodes and
                    not descriptor.allowed_on_nodes(central_cli_nodes)):
                continue
            # Add current descriptor to list of suggestions
            if descriptor._doc_override:
                result.add(descriptor._doc_override)
            else:
                cmd = []
                for item in path_tuple:
                    if is_argument(item):
                        arg_desc = self._get_argument_descriptor(item)
                        item = arg_desc.label
                        # Append CENTRAL_CLI_END_TOKEN for Central CLI commands
                        if arg_desc.id == 'RegisteredNodeUuid':
                            item = '%s %s' % (item, CENTRAL_CLI_END_TOKEN)
                    cmd.append(item)
                result.add(" ".join(cmd))

        return list(result)

    def get_command_tree(self, context=MODE_ENABLE):
        """Return python dict of dicts representing supported CLI commands.

        Args:
            context: Current context.
        Note:
            Supported commands:
            - have descriptors that don't specify the hidden property.
            - have descriptor.context_available properties that contain context.
        """
        tree = {}
        if context not in self._factory.commands:
            return {}
        items = list(self._factory.commands[context].items())
        for descriptor, cmd_list in items:
            if (descriptor.hidden or
                    not self._allow_command_descriptor(descriptor, context)):
                continue
            cmd_list_len = len(cmd_list)
            pos = 0
            node = tree
            for token in cmd_list:
                node = node.setdefault(token, {})
                pos += 1
                # If descriptor supports additional_last_args, account for it
                if descriptor.additional_last_args and pos == cmd_list_len:
                    for count in range(descriptor.additional_last_args):
                        node[END_TOKEN] = {}
                        node = node.setdefault(token, {})
            node[END_TOKEN] = {}
            if not descriptor.suppress_modifiers:
                node[OUTPUT_MODIFIER_TOKEN] = FORMATTERS
        return tree

    def _navigate_command(self, descriptor, context, cmd_list):
        """Return whatever descriptor is a navigation command and next context.

        Args:
            descriptor: CLICommand descriptor
            context: current CLI context
            cmd_list: List of cmd tokens specified
        Return:
            (is_navigation_command, next_context) where:
            - is_navigation_command specifies whether the specified
              descriptor is a navigation command.
            - next_context specifies the new context if the command is executed,
              or None if the command is not valid in the current context, for
              example, if a Central CLI context command is invoked when not in
              session mode.
        """
        if not isinstance(descriptor, CliCommandDescriptor):
            return False, context
        # Detemine next context based on specified descriptor
        context_data = None
        if descriptor.id == "Exit":
            next_context = self._PREVIOUS_CONTEXT.get(context, MODE_QUIT)
        elif descriptor.id == "Exit%s" % CENTRAL_CLI_STR:
            if not self.session_mode_active:
                return True, None
            elif self._session_context_prefix:
                # Exit context within a session
                self._session_context_prefix = ""
            else:
                # Exit session mode
                self._session_nodes = []
            next_context = MODE_ENABLE
        elif descriptor.context_command:
            if (is_central_cli_command(cmd_list) and
                    not self.session_mode_active):
                # Don't allow a "on <uuid> exec CMD" invocation if CMD is
                # a context command, we only allow context commands on remote
                # nodes while in session mode
                return True, None
            output = descriptor.invoke(cmd_list)
            parts = output.split("\n\n")
            if len(parts) != 3 or parts[0] != "OK":
                lg.error("Invalid context command result: %s", output,
                         error_code=ec.OPERATION_FAILED)
                return False, context
            if self._output_navigate_cmd:
                sys.stdout.write('%s\n' % output)
            next_context = simplejson.loads(parts[1])
            context_data = simplejson.loads(parts[2])
            if (self.session_mode_active and
                    descriptor.id.endswith(CENTRAL_CLI_STR)):
                self._session_context_prefix = " ".join(cmd_list)
        else:
            next_context = self._NEXT_CONTEXT.get(descriptor.id)
        # If specified descriptor is not a navigation command, return now
        if not next_context:
            return False, context
        # Specified descriptor is a navigation command, update the prompt
        elif next_context != MODE_QUIT and self._next_prompt_func:
            if isinstance(context_data, dict):
                self._next_prompt_func[next_context](**context_data)
            else:
                self._next_prompt_func[next_context]()
        # Update last known context and context_data
        self._context = next_context
        self._context_data = context_data
        if self._context in [MODE_INTERFACE]:
            self._context_data = cmd_list[1]
        return True, next_context

    def get_token_help(self, token):
        """Return command token help text for command completion queries.

        Args:
            token: A token within a command (e.g. "version" in "show version")
        Return:
            (token, help summary) tuple to display when the user queries
            command completion help (i.e. enters "show ?")
        Notes:
            - token can be an argument (e.g. <hostname>) in which case,
              the return token is the label of the argument as defined
              by its argument descriptor (e.g. HOSTNAME)
            - if the token is the END_TOKEN or the EXECUTE_COMMAND_HELP_ID
              (.e.g "<cr>"), the "Execute command" help summary is returned.
        """
        if token in [END_TOKEN, EXECUTE_COMMAND_HELP_ID]:
            return (EXECUTE_COMMAND_HELP_ID,
                    self._factory.help_summary.get(EXECUTE_COMMAND_HELP_ID, ""))
        elif is_argument(token):
            arg_desc = self._get_argument_descriptor(token)
            return arg_desc.label, arg_desc.help_summary
        return token, self._factory.help_summary.get(token, "")

    def is_argument(self, arg_name):
        """Return True if specified arg_name specified with argument indicators.

        Args:
            arg_name: Name of argument.
        Return:
            True, if arg_name has argument indicators, i.e. "<hostname>"
            returns True, "hostname" returns False.
        """
        return is_argument(arg_name)

    def _get_argument_descriptor(self, arg_name):
        """Return CLIArgument descriptor of specified argument name.

        Args:
            arg_name: Name of argument, can be "name" or "<name>".
        Return:
            A CLIArgument descriptor, on None if arg_name not found.
        """
        arg_name = resolve_argument_name(arg_name)
        return self._factory.arguments.get(arg_name)

    def _get_argument_values(self, arg_name, return_json=False):
        """Return valid values of specified argument name.

        Args:
            arg_name: Name of argument, can be "name" or "<name>".
            return_json: True, if return list in json format.
        Return:
            Valid values in list or json list format.  If arg_name not
            found, an empty list (optionally in json format) is returned.
        """
        arg_desc = self._get_argument_descriptor(arg_name)
        vals = arg_desc.get_allowed_values(
            self._req_dict, debug_mode=self._allow_debug_cmds,
            session_nodes=self.session_nodes) or [] if arg_desc else []
        vals = [force_unicode(val) for val in vals]
        return simplejson.dumps(vals) if return_json else vals

    def _get_matching_value(self, value, allowed_values, allow_multiple,
                            arg_desc):
        """Return matching argument value among possible allowed_values"""
        values = value.split()
        # values without ending_word
        values_wo_ew = list(set(allowed_values)-{arg_desc.ending_word})
        total_value_items = len(values)

        if arg_desc.full_match_required:
            # If argument is not multi_arg
            if arg_desc.limit == 1:
                return value if value in allowed_values else None

            has_ending_word = (values[-1] == arg_desc.ending_word)

            if has_ending_word:
                return (value
                        if all([v in values_wo_ew for v in values[:-1]]) else
                        None)
            else:  # Does not contain ending_word
                # All value in values are value tokens and equal to limit
                if (all([(v in values_wo_ew) for v in values]) and
                        len(values) == arg_desc.limit):
                    return value
                # All matching similar to limit, but total less than limit
                elif all([(v in values_wo_ew) for v in values]):
                    return "%s %s" % (value, arg_desc.ending_word)
                else:
                    return None
        else:  # Partial match is okay
            # All matches in allowed_values, which start with value
            if arg_desc.limit == 1:
                allowed_values.sort()
                matches = [
                    val for val in allowed_values if val.startswith(value)]
            else:  # multi_args
                completing_ending_word = arg_desc.ending_word.startswith(
                    values[-1])

                if completing_ending_word:
                    if len(values[:-1]) > arg_desc.limit:
                        return None
                    elif all([val in values_wo_ew for val in values[:-1]]):
                        matches = [
                            '%s %s' % (' '.join(values[:-1]),
                                       arg_desc.ending_word)]
                    else:
                        return None
                else:
                    if len(values) > arg_desc.limit:
                        return None
                    elif all([val in values_wo_ew for val in values[:-1]]):
                        val_matches = [
                            val for val in values_wo_ew
                            if val.startswith(values[-1])]
                        if len(values) == 1:
                            matches = [
                                '%s %s' % (val, arg_desc.ending_word)
                                for val in val_matches]
                        else:
                            matches = [
                                '%s %s %s' % (
                                    ' '.join(values[:-1]), val,
                                    arg_desc.ending_word) for val in val_matches]
                    else:
                        return None

            total_matches = len(matches)

            if total_matches < 1:
                return None
            elif total_matches == 1:
                return matches[0]
            elif not allow_multiple:
                return value if value in matches else None
            else:
                return matches[0]

    def match_argument_value(self, arg_name, value, allow_multiple=False,
                             verify_argument=True):
        """Return matched value if it matches argument specified by arg_name;
           else None.

        Notes:
            1. If value is a substring of one of the possible values, it is
               considered a match, unless full_match_required is enabled in
               the CliArgument descriptor.
               For example:
               - "fu" matches possible values ["foobar", "fubar"]
               - "foobar" matches possible values ["foobar", "fubar"]
               - "fubar" matches possible values ["foobar", "fubar"]

            2. If value is a substring of multiple possible values, what's
               returned depends on whether the argument's full_match_required
               flag is set and whether allow_multiple is True or False:

               For example if not full_match_required and allow_multiple=False:
               - "fu" matches possible values ["fu", "fubar", "full"]
                 (In this case this method returns "fu")
               - "fu" does not match possible values ["fubar", "full"]
                 (In this case this method returns None)

               For example if not full_match_required and allow_multiple=True:
               - "fu" matches possible values ["fu", "fubar", "full"]
                 (In this case this method returns "fu")
               - "fu" matches possible values ["fubar", "full"]
                 (In this case this method can return "fubar" or "full")

            3. Substring matches are not possible with pattern or verify args.

            4. verify_argument controls whether arguments of type verify_arg
               would actually perform verification. By default we'd always want
               this to be True, unless verification is done using API calls, eg
               in case of Central CLI commands. In that situation verification
               for each descriptor may result in extremely long response time.
               So by making verify_argument False for such cases, we can avoid
               that
        """
        arg_desc = self._get_argument_descriptor(arg_name)
        if not arg_desc:
            return None
        elif not arg_desc.verify:
            # return value as is if argument verification is disabled
            return value
        elif arg_desc.pattern_argument:
            return value if re.match(arg_desc._pattern, value) else None
        elif arg_desc.verify_argument:
            if not verify_argument:
                return value
            return (value
                    if arg_desc.match_verify_arg_value(
                        value, self._req_dict,
                        debug_mode=self._allow_debug_cmds,
                        remote_nodes=self.session_nodes\
                            if self.session_mode_active else None)
                    else None)
        # Get value for query argument from cache, and,
        # if not found then process the request & store back the values in cache.
        if arg_name in self._cached_remote_args:
            allowed_values = self._cached_remote_args[arg_name]
        else:
            allowed_values = self._get_argument_values(arg_name)
            self._cached_remote_args[arg_name] = allowed_values

        return self._get_matching_value(value, allowed_values,
                                        allow_multiple, arg_desc)

    def search_argument_values(self, arg_name, starts_with=""):
        """Search specified argument for values.

        Args:
            arg_name: Name of argument, can be "name" or "<name>".
            starts_with: Search filter to apply on values.
        Return:
            (is_argument, is_enum_argument, is_query_argument,
            values_found list) tuple where:
            - is_argument is True if arg_name is an argument.
            - is_enum_argument is True if arg_name is an enum argument.
            - is_query_argument is True if arg_name is a query argument.
            - values_found is a list (value, desciption) tuples where
              values are all argument vlaues that match the starts_with
              filter and description is the value description.
        """
        values = []
        arg_desc = self._get_argument_descriptor(arg_name)
        if not arg_desc:
            return False, False, False, values
        enum_arg = False
        query_arg = False
        if (arg_desc.enum_argument or arg_desc.query_argument):
            if arg_desc.enum_argument:
                enum_arg = True
            else:
                query_arg = True
            # Incase of query arg; fetch its respective values from cache, if found.
            if query_arg and arg_name in self._cached_remote_args:
                values = self._cached_remote_args[arg_name]
            else:
                for value in self._get_argument_values(arg_name):
                    if value.startswith(starts_with):
                        desc = arg_desc.get_argument_value_desc(
                            value, self._req_dict)
                        values.append((value, desc))
        elif arg_desc.verify_argument:
            if (starts_with and
                    arg_desc.match_verify_arg_value(
                        starts_with, self._req_dict,
                        debug_mode=self._allow_debug_cmds,
                        remote_nodes=self.session_nodes\
                            if self.session_mode_active else None)):
                desc = arg_desc.get_argument_value_desc(
                    starts_with, self._req_dict)
                values = [(starts_with, desc)]
        elif arg_desc.pattern_argument:
            if starts_with and re.match(arg_desc._pattern, starts_with):
                desc = arg_desc.get_argument_value_desc(
                    starts_with, self._req_dict)
                values = [(starts_with, desc)]

        # Incase of query arg, store back the value of arg if not present in cache.
        if query_arg and arg_name not in self._cached_remote_args:
            self._cached_remote_args[arg_name] = values

        return True, enum_arg, query_arg, values

    def _get_configuration_command_output(self, cmd_list, category_filter=None):
        """Get 'show configuration' output

        Args:
            cmd_list: List of cmd tokens specified
            category_filter: Show only configuration that match filter.
        Return:
            List of configuration commands to output.
        """
        def _add_to_config_results(result, descriptors, category):
            for descriptor in descriptors:
                # Ignore Central CLI descriptors
                if is_central_cli_command(descriptor.path_tuple):
                    continue
                config = descriptor.show_configuration()
                if not config:
                    continue
                config = simplejson.loads(config)
                if isinstance(config, compat23_string_types):
                    result.setdefault(category, []).append(config)
                if isinstance(config, list):
                    result.setdefault(category, []).extend(config)
                else:
                    assert False, ("Unexpected config type: %s, config: %s" %
                                   (type(config), str(config)))
                    continue

        result = {}
        show_configs = self._factory.show_config_commands
        if category_filter in [None, SHOW_CONFIG_DNS]:
            _add_to_config_results(
                result, show_configs[SHOW_CONFIG_DNS], SHOW_CONFIG_DNS)
        if category_filter in [None, SHOW_CONFIG_GLOBAL]:
            _add_to_config_results(
                result, show_configs[SHOW_CONFIG_GLOBAL], SHOW_CONFIG_GLOBAL)
        if category_filter in [None, SHOW_CONFIG_INTERFACE]:
            _add_to_config_results(
                result, show_configs[SHOW_CONFIG_INTERFACE],
                SHOW_CONFIG_INTERFACE)
        return result

    def _valid_show_config_category(self, cmd_list):
        """Return True if a valid 'show configuration' command specified."""
        if len(cmd_list) == 2:
            return True
        elif (len(cmd_list) == 3 and
              cmd_list[2] in
              [SHOW_CONFIG_DNS, SHOW_CONFIG_GLOBAL, SHOW_CONFIG_INTERFACE]):
            return True
        return False

    def get_context(self):
        """Return information about last known context.

        Return:
            (context, context_data) tuple where context can be one of the MODE_
            values defined in utils.constants and context_data is additional
            information associated with a context. For example when the user
            enters the interface context via "interface eth0", the value
            returned for context_data is eth0. context_data can also be None
            if no data is associated with a context, e.g. enable, configure,
            etc.
        """
        return self._context, self._context_data

    def set_context_data(self, new_data):
        """Set the context data.

        User can call this API to save any context data.
        new_data:
            new context data that caller want to save

        Return:
            (context, context_data) tuple
        """

        self._context_data = new_data
        return self._context, self._context_data

    def get_requires_restart(self):
        """Gets the status if nsx-cli restart is required.
        Some newly set system settings may not be available in current nsx-cli session,
        this flag specifies whether restart of nsx-cli is required for these settings
        to be in effect.

        Return:
            bool: (requires_restart)
        """
        return self._requires_restart

    def set_requires_restart(self):
        """Sets the status of nsx-cli to restart is required.
        Some newly set system settings may not be available in current nsx-cli session,
        this flag specifies whether restart of nsx-cli is required for these settings
        to be in effect.
        """
        self._requires_restart = True

    def get_modifier(self):
        """Return modifier string specified for current command.

        Return:
            Modifier string specified by user for the current command, or
            None, if the user did not specify a modifier.
        """
        return self._modifier_str

    def set_req_dict(self, req_dict):
        """Used by utils/input_utils.py to configure req_dict when the user
           invokes the tab completion feature."""
        self._req_dict = req_dict

    def process_stdin_arg_request(self, input_dict):
        """Process argument requests made through stdin input"""
        # Verify input data
        input_type = input_dict.get("input_type")
        if input_type not in ["get_arg_values", "verify_arg"]:
            lg.error(
                "Invalid stdin input, unknown input_type: %s", input_type,
                error_code=ec.INVALID_ARGUMENT)
            return STDIN_INPUT_INVALID_REQUEST_ERROR, None
        arg_id = input_dict.get("arg_id")
        command = input_dict.get("command")
        context = input_dict.get("context")
        value = input_dict.get("value")
        if (arg_id is None or command is None or context is None or
                (input_type == "verify_arg" and value is None)):
            lg.error("Invalid stdin input, missing arg_id, command, "
                     "context or value", error_code=ec.INVALID_ARGUMENT)
            return STDIN_INPUT_MISSING_INPUT_ERROR, None
        # Verify valid argument id
        arg_desc = self._factory._descriptors.get(arg_id)
        if not isinstance(arg_desc, CliArgumentDescriptor):
            lg.error("Specified stdin arg_id invalid: %s", arg_id,
                     error_code=ec.INVALID_ARGUMENT)
            return STDIN_INPUT_BAD_ARG_ID_ERROR, None
        # Process request
        output = None
        req_dict = {
            "cmd_list": command.split(),
            "context": context}
        if input_type == "get_arg_values":
            results = {}
            values = arg_desc.get_allowed_values(
                req_dict, debug_mode=self._allow_debug_cmds,
                session_nodes=self.session_nodes)
            for val in values:
                if value and not val.startswith(value):
                    continue
                desc = arg_desc.get_argument_value_desc(val, req_dict)
                results[val] = desc
            if results:
                output = []
                for key in sorted(results):
                    output.append("%s\t%s" % (key,  results[key]))
                output = "\n".join(output)
        elif input_type == "verify_arg":
            if not arg_desc.verify_argument:
                lg.error("Specified stdin verify arg_id invalid: %s", arg_id,
                         error_code=ec.INVALID_ARGUMENT)
                return STDIN_INPUT_BAD_VERIFY_ARG_ID_ERROR, None
            elif not arg_desc.match_verify_arg_value(
                    value, req_dict, debug_mode=self._allow_debug_cmds,
                    remote_nodes=self.session_nodes\
                            if self.session_mode_active else None):
                return STDIN_INPUT_BAD_VERIFY_ARG_VALUE_ERROR, None
        return 0, output
