# ***************************************************************************
# Copyright 2017-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
import gzip
from logging import FileHandler
from logging import StreamHandler
from logging.handlers import BaseRotatingHandler
from logging.handlers import RotatingFileHandler
from logging.handlers import TimedRotatingFileHandler
import os
import sys
import time
if sys.platform != "win32":
    import stat


class FileChownHandler(FileHandler):
    """
    This file handler is used to restore log file owner uid and gid.
    """

    def __init__(self, usrname=None, grpname=None, *args, **kwargs):
        try:
            import grp
            import pwd
            self.uid = pwd.getpwnam(usrname).pw_uid if usrname else None
            self.gid = grp.getgrnam(grpname).gr_gid if grpname else None
        except:
            self.uid = None
            self.gid = None
        super(FileChownHandler, self).__init__(*args, **kwargs)

    def _open(self):
        stream = super(FileChownHandler, self)._open()
        try:
            if sys.platform != "win32":
                if self.gid:
                    os.chown(self.baseFilename, -1, self.gid)
                if self.uid:
                    os.chown(self.baseFilename, self.uid, -1)
                os.chmod(
                    self.baseFilename,
                    stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP)
        except:
            pass
        return stream


class RegularFileHandler(FileHandler):
    """
    This file handler is based on logging.FileHandler.
    Added an interface to reopen the logging file, which is used to support
    system log rotate.

    Notice that developer have to specify filename for the log.
    """

    def reopen(self):
        StreamHandler.__init__(self, self._open())


class MultiProcessRotatingHandler(BaseRotatingHandler):
    """
    This class provided a 'wrapper' emit method which handles the multi-process
    scenario. Make sure that other classes in self's mro doesn't have anything
    to do with emit.
    """

    def _lock_closure(self):
        """
        This method will return a pair of lock and unlock method.
        Currently file lock is not supported on windows.
        """
        def _lock_posix():
            try:
                with open(self.lockname, "w") as lock_fd:
                    fcntl.flock(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                return True
            except:
                return False
            return False

        def _unlock_posix():
            with open(self.lockname, "w") as lock_fd:
                fcntl.flock(lock_fd, fcntl.LOCK_UN)

        def _lock_win():
            return True

        def _unlock_win():
            pass

        try:
            import fcntl
            return _lock_posix, _unlock_posix
        except ImportError:
            return _lock_win, _unlock_win

    def __init__(self, lockname, pre_emit_callback=None, *args, **kwargs):
        """
        Just take the parameter 'lockname', which is a filename used by
        multiple processes as a filelock.
        """
        self.lockname = lockname
        self._pre_emit_callback = None
        if pre_emit_callback is not None:
            self._pre_emit_callback = pre_emit_callback
        self.flock, self.funlock = self._lock_closure()
        super(MultiProcessRotatingHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        # some constants
        ATTEMPTS = 100
        SLEEP_DURATION = 0.05

        # update the callback function if exists
        callback = None
        if hasattr(self, "_pre_emit_callback"):
            callback = self._pre_emit_callback

        # acquire lock if in multiple processes
        try:
            acquired, attempts = False, ATTEMPTS
            while attempts > 0:
                acquired = self.flock()
                if acquired:
                    break
                else:
                    time.sleep(SLEEP_DURATION)
                    attempts -= 1
            if acquired:
                if callback is not None:
                    callback()
                super(MultiProcessRotatingHandler, self).emit(record)
        except:
            pass

        # clean up
        finally:
            try:
                if self.stream:
                    self.close()
                if acquired:
                    self.funlock()
            except:
                pass


class CompressedRotatingHandler(BaseRotatingHandler):
    """
    Provide functions to perform compress and
    """

    def __init__(self, compress=None, greenlet=False, *args, **kwargs):
        self.greenlet = greenlet
        self.compress = None
        # currently only support '.gz' compress format
        if compress is not None:
            self.compress = "gz"
        self.spawn = self._multi_task_closure()
        super(CompressedRotatingHandler, self).__init__(*args, **kwargs)

    def _multi_task_closure(self):
        """
        config a helper method, which can spawn a file compression task,
        use different lib based on user config and runtime environment.
        """
        # multi-task closure
        # TODO: for now invoke synchronously; otherwise, rotation for
        # single or multiple processes (with lockname) fail
        def use_gevent(func, *args, **kwargs):
            spawn(func, *args, **kwargs).join()

        def use_threading(func, *args, **kwargs):
            t = Thread(target=func, args=args, kwargs=kwargs)
            t.start()
            t.join()

        def use_multiprocessing(func, *args, **kwargs):
            p = Process(target=func, args=args, kwargs=kwargs)
            p.start()
            p.join()

        if sys.platform == "win32":
            from threading import Thread
            return use_threading
        elif self.greenlet is True:
            try:
                from gevent import spawn
                return use_gevent
            except ImportError:
                from threading import Thread
                return use_threading
        else:
            from multiprocessing import Process
            return use_multiprocessing

    def doRollover(self):
        def replace_file(sfn, dfn):
            if os.path.exists(dfn):
                os.remove(dfn)
            os.rename(sfn, dfn)

        def compress_file(sfn, dfn, gid, ext):
            gid = int(gid)
            gzip_file = None
            compress_error = False
            try:
                gzip_file = gzip.open(dfn + ext, "wb")
                with open(sfn) as fp:
                    while True:
                        chunk = fp.read(1024)
                        if not chunk:
                            break
                        gzip_file.write(chunk.encode("utf-8"))
                        # implicit context switch
                        time.sleep(0)
            except Exception:
                compress_error = True
            finally:
                try:
                    if gzip_file:
                        gzip_file.close()
                        if compress_error:
                            os.remove(gzip_file.filename)
                        elif sys.platform != "win32":
                            if gid != -1:
                                os.chown(gzip_file.filename, -1, gid)
                            os.chmod(
                                gzip_file.filename,
                                stat.S_IRUSR | stat.S_IWUSR |
                                stat.S_IRGRP | stat.S_IWGRP)
                    if not compress_error:
                        os.remove(sfn)
                except Exception:
                    pass

        if self.stream:
            self.stream.close()
            self.stream = None
        gid = self.gid if hasattr(self, "gid") and self.gid else -1
        if self.backupCount > 0:
            ext = "" if self.compress is None else ("." + self.compress)
            for i in range(self.backupCount - 1, 0, -1):
                sfn = "%s.%d" % (self.baseFilename, i)
                dfn = "%s.%d" % (self.baseFilename, i + 1)
                if os.path.exists(sfn):
                    if not ext:
                        replace_file(sfn, dfn)
                        continue
                    self.spawn(compress_file, sfn, dfn, gid, ext)
                elif ext != "" and os.path.exists(sfn + ext):
                    replace_file(sfn + ext, dfn + ext)

            # add basefile to compress list
            dfn = self.baseFilename + ".1"
            if not ext:
                replace_file(self.baseFilename, dfn)
            else:
                self.spawn(compress_file, self.baseFilename, dfn, gid, ext)
        if not self.delay:
            self.stream = self._open()


class CombinedRotatingFileHandler(TimedRotatingFileHandler):
    """
    Combined rotating file handler based on both size and time.

    The log records that exceed maxBytes will be ignored until the time for
    rotation.
    """

    def __init__(self, maxBytes, *args, **kwargs):
        self.maxBytes = maxBytes
        super(CombinedRotatingFileHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        if self.maxBytes and self.stream.tell() > self.maxBytes:
            return
        super(CombinedRotatingFileHandler, self).emit(self, record)


def GetRotatingFileHandler(**kwargs):
    """
    Return a configured rotating log file handler.

    Notice the result handler doesn't contain filter or formatter, and need to
    be set manually. Also, these arguments (formatter, filter)should not
    appear in kwargs.
    """

    def _is_path_valid(path):
        return isinstance(path, str)

    if "filename" not in kwargs:
        return None

    class_name, class_obj = "", None
    daemon = kwargs.pop("daemon", False)
    if daemon:
        class_obj = RegularFileHandler
        class_name = "RegularFileHandler"
    else:
        multi, compress, size_based, time_based, fchown = (
            False, False, False, False, False)

        if "lockname" in kwargs and _is_path_valid(kwargs["lockname"]):
            multi = True
        if "compress" in kwargs and isinstance(kwargs["compress"], str):
            compress = True
        if "maxBytes" in kwargs and isinstance(kwargs["maxBytes"], int):
            size_based = True
        # for time based, our policy is developer need to specify the time for
        # rotation instead of using the default
        if "when" in kwargs and isinstance(kwargs["when"], str):
            time_based = True
        if "usrname" in kwargs or "grpname" in kwargs:
            fchown = True

        file_handler, file_handler_name = None, ""
        if size_based and time_based:
            file_handler = CombinedRotatingFileHandler
            file_handler_name = "Combined"
        elif not time_based:
            file_handler = RotatingFileHandler
            file_handler_name = "SizeBased"
        else:
            file_handler = TimedRotatingFileHandler
            file_handler_name = "TimeBased"

        base_classes, class_name = (), ""
        if fchown:
            base_classes += (FileChownHandler,)
            class_name += "Chown"
        if multi:
            base_classes += (MultiProcessRotatingHandler,)
            class_name += "MultiProcess"
        if compress:
            base_classes += (CompressedRotatingHandler,)
            class_name += "Compressed"
        base_classes += (file_handler,)
        class_name += (file_handler_name + "RotatingFileHandler")
        class_obj = type(class_name, base_classes, {})
    return class_obj(**kwargs)
