# Copyright (C) 2021 VMware, Inc. All rights reserved.
import socket
import sys
import platform
import time

try:
    SYSTEM = platform.system()
    if SYSTEM == "VMkernel":
        sys.path.append('/usr/lib/vmware/netopa/lib/python')
        sys.path.append('/usr/lib/vmware/nsx-common/lib/python')
        sys.path.append('/usr/lib/vmware/nsx-proto2/python/')
        from vmware.nsx.sha import sha_pm_pb2
    elif SYSTEM == "Linux":
        sys.path.append('/opt/vmware/nsx-netopa/lib/python')
        from vmware.nsx.sha import sha_pm_ts_pb2 as sha_pm_pb2
    # Need not remediation on windows so far as watchdog keep trying to start crashed daemon.
    # elif SYSTEM == "Windows":
    #     # resolving NSX installation dir
    #     import winreg
    #
    #     nsx_reg_path = r'SOFTWARE\VMWARE, Inc.\NSX'
    #     hKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, nsx_reg_path)
    #     nsx_path, regtype = winreg.QueryValueEx(hKey, 'InstallDir')
    #     if regtype != winreg.REG_SZ:
    #         raise TypeError(
    #             'NSX installation dir registry is not a string, current value: {}'.format(nsx_path))
    #
    #     proto_path = os.path.join(nsx_path, 'nsx-netopa', 'lib', 'python')
    #     sys.path.append(proto_path)
    else:
        raise Exception("Unsupported system:[{}]".format(SYSTEM))
except Exception as e:
    print("Failed in setting sys path. Error:{}".format(e))

SOCKET_PATH = "/var/run/vmware/nsx-sha/sha_socket"
HOST = '127.0.0.1'  # SHA server's IP address
PORT = 11111  # Port used by SHA server
DELIMITER = 'SHACLIENT'
NAME_SPACE = 'nestdb-init'
SERVER_REPLY = 'bumping'
UTF_8 = 'utf-8'


def create_start_msg():
    client_msg = sha_pm_pb2.ShaClientMsg()
    client_msg.name_space = NAME_SPACE
    item_msg = client_msg.item.add()
    item_msg.name_space = NAME_SPACE
    item_msg.key = 'start'
    item_msg.timestamp = int(time.time() * 1000)
    return client_msg.SerializeToString()


def create_stop_msg():
    client_msg = sha_pm_pb2.ShaClientMsg()
    client_msg.name_space = NAME_SPACE
    item_msg = client_msg.item.add()
    item_msg.name_space = NAME_SPACE
    item_msg.key = 'stop'
    item_msg.timestamp = int(time.time() * 1000)
    return client_msg.SerializeToString()


def create_start_by_init_msg():
    client_msg = sha_pm_pb2.ShaClientMsg()
    client_msg.name_space = NAME_SPACE
    item_msg = client_msg.item.add()
    item_msg.name_space = NAME_SPACE
    item_msg.key = 'nestdb_is_started'
    item_msg.boolean_value = True
    item_msg.type = sha_pm_pb2.ShaClientItemMsg.ItemMsgType.Value('BOOLEAN')
    item_msg.timestamp = int(time.time() * 1000)
    return client_msg.SerializeToString()


def create_stop_by_init_msg():
    client_msg = sha_pm_pb2.ShaClientMsg()
    client_msg.name_space = NAME_SPACE
    item_msg = client_msg.item.add()
    item_msg.name_space = NAME_SPACE
    item_msg.key = 'nestdb_is_started'
    item_msg.boolean_value = False
    item_msg.type = sha_pm_pb2.ShaClientItemMsg.ItemMsgType.Value('BOOLEAN')
    item_msg.timestamp = int(time.time() * 1000)
    return client_msg.SerializeToString()


def send_msg(sock, serial_msg):
    final_msg = DELIMITER.encode(UTF_8) + get_len_array(serial_msg) + serial_msg
    sock.sendall(final_msg)


def get_len_array(serial_msg):
    return len(serial_msg).to_bytes(4, byteorder='big')


def connect_sock():
    if SYSTEM == "Linux":
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(SOCKET_PATH)
    else:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((HOST, PORT))
    return sock


def send_by_init(op_type):
    retry = 3
    err_msg = ''
    while retry > 0:
        sock = None
        try:
            sock = connect_sock()
            start_msg = create_start_msg()
            send_msg(sock, start_msg)
            check_reply(sock)
            if op_type == 'start_by_init_script':
                msg = create_start_by_init_msg()
            elif op_type == 'stop_by_init_script':
                msg = create_stop_by_init_msg()
            else:
                raise Exception('Invalid argument:[{}]'.format(op_type))
            send_msg(sock, msg)
            check_reply(sock)
            stop_msg = create_stop_msg()
            send_msg(sock, stop_msg)
            check_reply(sock)
            print("Succeed")
            return
        except Exception as e:
            err_msg = "Failed in sending msg for op_type:[{}], error:{}".format(op_type, e)
            retry -= 1
        finally:
            if sock is not None:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()

    print(err_msg)


def check_reply(sock):
    resp = receive_msg(sock, len(SERVER_REPLY)).decode(UTF_8)
    if resp != SERVER_REPLY:
        raise Exception("Not a valid response:[{}].".format(resp))


def receive_msg(sock, msg_len):
    chunks = []
    bytes_recd = 0
    while bytes_recd < msg_len:
        chunk = sock.recv(min(msg_len - bytes_recd, 2048))
        if chunk == b'':
            raise RuntimeError("socket connection broken")
        chunks.append(chunk)
        bytes_recd = bytes_recd + len(chunk)
    return b''.join(chunks)


op_type = sys.argv[1]
send_by_init(op_type)

