# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import re
import sha.config
import traceback
import xml.etree.ElementTree as ET
import uuid
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache


_APPLIANCE_INFO_XML = "/etc/vmware/nsx/appliance-info.xml"
_LOOKUP_CMD = ["sudo", "/usr/bin/nslookup"]


class ManagerFqdnReverseLookupFailureStatus(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._status_dict = {}

    def _get_manager_info(self):
        manager_info = []
        if os.path.isfile(_APPLIANCE_INFO_XML):
            tree = ET.parse(_APPLIANCE_INFO_XML)
            for proxy in tree.iter(tag = "appliance-proxy"):
                fqdn = proxy.find("fqdn")
                manager_info.append({
                    "ip": proxy.find("ip").text,
                    "fqdn": fqdn.text if fqdn is not None else "",
                    "uuid": proxy.find("uuid").text
                })
        return manager_info

    def _get_eth0_ip(self):
        try:
            output = run_command(['ifconfig', 'eth0'])
            inet_match = re.findall('.*inet.*',output)
            # Expected inet line:
            # inet aaa.bbb.ccc.ddd  netmask aaa.bbb.ccc.ddd  broadcast aaa.bbb.ccc.ddd
            ips = re.match('.*inet ([\d\.]+)\W+netmask.*', inet_match[0])
            return ips.groups()[0]
        except Exception:
            self._logger.warning("Failed to get eth0 IP address with error "
                "reason: %s", traceback.format_exc())

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        manager_info = self._get_manager_info()
        node_ip = self._get_eth0_ip()
        self._logger.debug("ManagerFqdnReverseLookupFailureStatus: "
            "manager_info = %s, node_ip = %s", str(manager_info), node_ip)
        status_dict = {}
        for info in manager_info:
            # Only do reverse lookup on local IP address
            if info["ip"] != node_ip:
                continue
            node_uuid = uuid.UUID(info["uuid"])
            node_fqdn = info["fqdn"] or ""
            context = {
                "entity_id": info["uuid"],
                "appliance_address": node_ip
            }
            status_dict[node_uuid] = (False, context)
            if node_fqdn and node_fqdn != node_ip:
                try:
                    out = run_command(_LOOKUP_CMD + [node_ip]) or ''
                    self._logger.debug("ManagerFqdnReverseLookupFailureStatus: "
                        "lookup output for %s is %s", str(node_ip), str(out))
                except Exception as e:
                    self._logger.warning("Reverse DNS lookup on manager IP {}"
                        " failed with error reason: %s", node_ip,
                        traceback.format_exc())
                    out = None
                # Set status to True if lookup failed or if the FQDN returned
                # does not contain the FQDN found in appliance-info.xml
                if not out or node_fqdn not in out.lower():
                    status_dict[node_uuid] = (True, context)
        # Delete old manager UUID in cases where node UUID changed
        deleted_entity_list =\
            list(set(self._status_dict.keys()) - set(status_dict.keys()))
        # Emit a log when there is a status change or entity id change
        if status_dict != self._status_dict or deleted_entity_list:
            self._logger.info("ManagerFqdnReverseLookupFailureStatus: "
                "status_dict = %s, deleted_entity_list = %s",
                str(status_dict), str(deleted_entity_list))
        self._status_dict = status_dict
        return self._status_dict, deleted_entity_list
