# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import time
import traceback
from threading import current_thread

import sha.config
from sha.forkserver import ForkServer


_logger = sha.config.get_logger("run_command")

# for PR 2536824
# TIME_THRESHOLD: the time threshold since the last cmd which did not
# encounter [Errno 28] (No space left on device) in execution.
#
# If current cmd encountered [Errno 28] in execution, and current timestamp
# is more than TIME_THRESHOLD since last cmd without [Error 28], to prevent
# SHA from being blocked by [Error 28], we should let SHA exit and restart
# by watchdog
TIME_THRESHOLD = sha.config.TIME_THRESHOLD

# Timestamp for last cmd execution without ([Errno 28] No space left on device)
_last_time_without_error28 = time.monotonic()

# input: only binary sequence allowed now
# custom: flag to mark the caller is custom plugins
def run_command(cmd, input=None, shell=False, timeout=None, custom=False, env=None):
    global _last_time_without_error28
    try:
        _logger.debug("run_command - start cmd:{}.".format(cmd))
        output = ForkServer.check_output(
            cmd,
            input=input,
            shell=shell,
            timeout=timeout,
            custom=custom,
            env=env
        ).decode('utf-8')
        # reset if cmd succeeds
        _last_time_without_error28 = time.monotonic()
        _logger.debug(
            "run_command - finished cmd:{} with output:{}.".format(cmd, output))
        return output
    except OSError as e:
        if e.errno != 28:
            # reset if not failed by [Errno 28]; raise as before
            _last_time_without_error28 = time.monotonic()
            raise e
        since_last = time.monotonic() - _last_time_without_error28
        if since_last < TIME_THRESHOLD:
            # time (since last execution without error28) not exceed threshold, raise as before
            _logger.error('Encountering [Errno 28], time since '
                          'last execution without error28 is :{} second'.format(since_last))
            raise e
        # [Errno 28] No space left on device,
        # continuous encountered above Error more than TIME_THRESHOLD,
        # SHA main process should exit and be restarted by watchdog
        trace = traceback.format_exc()
        _logger.error(
            'Exit SHA process as continuously encountering OSError - '
            '[Errno 28] No space left on device, trace:{}'.format(trace))
        import signal
        os.kill(os.getpid(), signal.SIGTERM)
