# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import nsx.lib.libdpcli as libvswitch
from nsx.exceptions import DPException, DvsFailure, VmkernelFailure
from vmware import vsi

import sha.config
from sha.contrib.monitor.uplinkstatus.esx.message import *

UPLINK_STATUS_UP_STR = 'OK'
UPLINK_STATUS_DOWN_STR = 'DOWN'

VMK_LINK_STATE_DOWN = 0
VMK_LINK_STATE_UP = 1

VSWITCH_TYPE = "vswitch"

UPLINK_DOWN_ALARM = "UPLINK_DOWN_ALARM"
LAG_MEMBER_DOWN_ALARM = "LAG_MEMBER_DOWN_ALARM"

_logger = sha.config.get_logger(__name__)


def set_uplink_down_alarm(status_changes):
    _logger.debug('set_uplink_down_alarm')
    status_changes[UPLINK_DOWN_ALARM] = True
    pass


def clear_uplink_down_alarm(status_changes):
    _logger.debug('clear_uplink_down_alarm')
    status_changes[UPLINK_DOWN_ALARM] = False
    pass


def set_lag_member_link_down_alarm(status_changes):
    _logger.debug('set_lag_member_link_down_alarm')
    status_changes[LAG_MEMBER_DOWN_ALARM] = True
    pass


def clear_lag_member_link_down_alarm(status_changes):
    _logger.debug('clear_lag_member_link_down_alarm')
    status_changes[LAG_MEMBER_DOWN_ALARM] = False
    pass


def is_portset_all_uplinks_down(vds_name):
    uplinks = libvswitch.GetUplinks(vds_name)
    if uplinks is None or not uplinks:
       return False
    for uplink in uplinks:
        if libvswitch.GetUplinkLinkState(vds_name, uplink) == UPLINK_STATUS_UP_STR:
            return False
    return True


def is_lag_member_link_down(event):
    return event == KCP_SHA_LACP_LINK_STATUS_EVENT_LINK_DOWN


def get_nsx_portset_names():
    """Get names of nsx portsets
    """
    portsets = vsi.list('/net/portsets')
    if portsets is None or not portsets:
        return []
    nsx_portsets = []
    for ps in portsets:
        properties_path = '/net/portsets/{0}/properties'.format(ps)
        properties = vsi.get(properties_path)
        if properties is None:
            continue
        switch_type = properties.get('implClass', None)
        if switch_type != VSWITCH_TYPE:
            continue
        nsx_portsets.append(ps)
    return nsx_portsets


def get_lag_table(portset):
    res = libvswitch.VSVMK_GetLAGTable(portset)
    if res[0]:
        return {}
    lag_table = res[1]
    lag_num = lag_table.numLAGs
    new_lag_table = {}
    cur_lag = None
    for i in range(lag_num):
        if i == 0:
            cur_lag = lag_table.vs_lag_table_first_lag()
        else:
            cur_lag = cur_lag.vs_lag_table_next_lag()
        new_lag = {}
        new_lag['name'] = cur_lag.get_name()
        new_lag['id'] = cur_lag.id
        new_lag['state'] = cur_lag.state
        new_lag['uplinkLinkupBitmap'] = cur_lag.uplinkLinkupBitmap
        new_lag['numUplinks'] = cur_lag.numUplinks
        new_lag['uplinkNames'] = list()
        for link_index in range(cur_lag.numUplinks):
            new_lag['uplinkNames'].append(cur_lag.get_uplink_name(link_index))
        new_lag_table[cur_lag.get_name()] = new_lag
    return new_lag_table


def is_dvs_all_uplinks_up(portset):
    """
    Check if all uplinks of the portset are up
    1. Get all Uplinks status per dev_name
    2. Get lag table
    3. Remove Lag member link from UplinkStatus table and add lag status
    4. Check if all uplinks status are up
    {"vmnic1": True}
    :param portset:
    :return:
    """
    uplink_status_map = dict()
    uplink_portid_status_map = dict()
    dvs_name = libvswitch.GetAliasNameFromPortset(portset)
    if dvs_name is None or dvs_name=='':
        _logger.warning('Failed to get dvs alias of portset {}'.format(portset))
        return True
    uplinks = libvswitch.GetUplinks(dvs_name)  # ['vmnic1', 'vmnic5', 'vmnic4']
    if not uplinks:
        return True
    for uplink in uplinks:
        link_state_up = True if libvswitch.GetUplinkLinkState(dvs_name, uplink) == UPLINK_STATUS_UP_STR else False
        uplink_status_map[uplink] = link_state_up
        port_id = libvswitch.GetUplinkDVPortID(uplink)
        uplink_portid_status_map[port_id] = link_state_up
    """
    {'ids': ['c4763fdd-f3a0-4346-9529-ac15028215ca', 'c803e18f-315f-4f76-886e-5f1f5b8c6b72',
             '2d9e2d15-1bd0-46fa-95d9-4ab908fbc5d4', '71470c38-ca8a-4c40-8dfd-201d0cf1a6b3',
             '7fc15c2c-58a3-45e3-8059-ee3380445b68'],
     'names': ['lag1-0', 'lag1-1', 'uplink1', 'uplink2', 'uplink3']
    }
    """
    uplink_ports = libvswitch.GetSwitchUplinkPorts(dvs_name)
    uplink_ports_names = uplink_ports['names']
    uplink_ports_ids = uplink_ports['ids']
    uplink_portname_id_map = dict()
    for i in range(len(uplink_ports_names)):
        uplink_portname_id_map[uplink_ports_names[i]] = uplink_ports_ids[i]
    """
    {
        "percentError": 0,
        "flags": ["SHOTGUN", "BEST_EFFORT", "NOTIFY_SWITCH"],
        "linkCriteria": ["DRV_PRESENT", "LINK_STATE"],
        "fullDuplex": TRUE,
        "activeUplinkPorts": ["uplink1"],
        "speed": 10,
        "standbyUplinkPorts": [],
        "loadBalance": "FO_EXPLICIT"
    }
    """
    active_uplink_ports = set()
    standby_uplink_ports = set()
    try:
        teaming_policy = libvswitch.GetTeamingPolicy(dvs_name)
        active_uplink_ports.update(set(teaming_policy['activeUplinkPorts']))
        standby_uplink_ports.update(set(teaming_policy['standbyUplinkPorts']))
    except VmkernelFailure:
        # teaming_policy not found
        pass
    try:
        ls_teaming_poicy_dict_list = libvswitch.GetLSTeamingPolicy(dvs_name)
        for ls_teaming_policy in ls_teaming_poicy_dict_list:
            active_uplink_ports.update(set(ls_teaming_policy['activeUplinkPorts']))
            standby_uplink_ports.update(set(ls_teaming_policy['standbyUplinkPorts']))
    except DvsFailure:
        # ls teaming policy not found
        pass
    # Get LAG state
    lag_table = get_lag_table(portset)
    is_all_uplinks_up = True
    for cur_uplink in set.union(active_uplink_ports, standby_uplink_ports):
        # normal uplink
        if cur_uplink in uplink_ports_names:
            if not uplink_portid_status_map.get(uplink_portname_id_map[cur_uplink], True):
                is_all_uplinks_up = False
                break
        else:
            # LAG state
            if cur_uplink not in lag_table:
                # TODO: _logger.debug lag not found error
                pass
            lag = lag_table[cur_uplink]
            if lag['state'] == VMK_LINK_STATE_DOWN:
                is_all_uplinks_up = False
                break
            for vmnic in lag['uplinkNames']:
                if not uplink_status_map[vmnic]:
                    is_all_uplinks_up = False
                    break
    return is_all_uplinks_up


def is_dvs_all_lag_member_links_up(portset):
    lag_table = get_lag_table(portset)
    is_all_lag_member_link_up = True
    for lag in lag_table.values():
        if lag['numUplinks'] <= 0:
            continue
        uplinkLinkupBitmap = lag['uplinkLinkupBitmap']
        if uplinkLinkupBitmap != (pow(2, lag['numUplinks']) - 1):
            is_all_lag_member_link_up = False
            break
    return is_all_lag_member_link_up


def handle_uplink_status_notification(notify, status_changes):
    if notify.event == KCP_SHA_LAG_STATUS_EVENT_DOWN:
        set_uplink_down_alarm(status_changes)
        return
    nsx_portsets = get_nsx_portset_names()
    for ps in nsx_portsets:
        if not is_dvs_all_uplinks_up(ps):
            set_uplink_down_alarm(status_changes)
            return
    clear_uplink_down_alarm(status_changes)


def handle_lag_link_status_notification(notify, status_changes):
    if notify.event == KCP_SHA_LACP_LINK_STATUS_EVENT_LINK_DOWN:
        set_lag_member_link_down_alarm(status_changes)
        return
    if notify.event == KCP_SHA_LAG_STATUS_EVENT_DOWN or \
        notify.event == KCP_SHA_LAG_STATUS_EVENT_UP:
        handle_uplink_status_notification(notify, status_changes)
        return
    nsx_portsets = get_nsx_portset_names()
    for ps in nsx_portsets:
        if not is_dvs_all_lag_member_links_up(ps):
            set_lag_member_link_down_alarm(status_changes)
            return
    clear_lag_member_link_down_alarm(status_changes)


def handle_init_notification(status_changes):
    nsx_portsets = get_nsx_portset_names()
    uplink_down = False
    lag_down = False

    for ps in nsx_portsets:
        if not uplink_down and not is_dvs_all_uplinks_up(ps):
            set_uplink_down_alarm(status_changes)
            uplink_down = True
        if not lag_down and not is_dvs_all_lag_member_links_up(ps):
            set_lag_member_link_down_alarm(status_changes)
            lag_down = True
        if uplink_down and lag_down:
            break

    if not uplink_down:
        clear_uplink_down_alarm(status_changes)
    if not lag_down:
        clear_lag_member_link_down_alarm(status_changes)


def handle_notification(notify, status_changes):
    # DVS config may be changed before receiving the nofication
    # Need recall DVSLoadConfig to refresh info in libvswitch
    #
    # Please note: It is safe to load config with lock free,
    # due to only one thread in EsxUplinkStatusMonitor is
    # using libvswitch now.
    if isinstance(notify, UplinkStatusNotification):
        libvswitch.DVSLoadConfig()
        handle_uplink_status_notification(notify, status_changes)
    elif isinstance(notify, LagLinkStatusNotification):
        libvswitch.DVSLoadConfig()
        handle_lag_link_status_notification(notify, status_changes)
    elif isinstance(notify, InitNotification):
        libvswitch.DVSLoadConfig()
        handle_init_notification(status_changes)
    else:
        _logger.error('handle_notification - unexpected notification:{}'.format(notify))
