# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import sha.config
import os
import re


class Context:
    """Secured context which will provide necessary access rights for plugins"""

    def __init__(self, name, paths=[]):
        self._logger = sha.config.get_logger(name)
        self._opened_file = {}
        self._local_map = {}
        self._name = name
        self._paths_reg = [re.compile(path) for path in paths]

    def get_logger(self):
        """Specific logger instance for each plugin"""
        return self._logger

    def get_local_map(self):
        """Since every plugin will be triggered when new metric data is ready,
        it can choose to store some stateful data in this store"""
        return self._local_map

    def open_readonly_file(self, path):
        """Plugins are not allowed to open any file in restricted environment.
        In order to fetch some data, it can use this method to read some non-critical
        files."""
        abspath = os.path.abspath(path)
        for reg in self._paths_reg:
            if reg.match(abspath) is not None:
                if abspath not in self._opened_file:
                    try:
                        file = open(abspath, "r")
                        self._opened_file[abspath] = file
                        return file
                    except IOError:
                        self._logger.error("Fail to open file: " + abspath)
                        return None
                else:
                    return self._opened_file[abspath]

        self._logger.warning("Plugins have no access to file: " + abspath)
        return None

    def close_all_file(self):
        """Close all opened files to avoid resource leak"""
        for file in self._opened_file.values():
            if not file.closed:
                try:
                    file.close()
                except IOError:
                    self._logger.error("Fail to close file: " + file.name)
        self._opened_file.clear()
