# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import sha.config


import ctypes
import re
import socket
import sys
import time

def sanitize(string):
    """Sanitize a string, replace whitespace with "-".
    @param string: Input string
    @return: Sanitized string
    """
    whitespace_sanitized = re.sub(r'[\s]+', '-', string)
    if '"' in whitespace_sanitized:
        return '"' + re.sub(r'[\"]+', '\\\\\"', whitespace_sanitized) + '"'
    return '"' + whitespace_sanitized + '"'

def is_blank(string):
    """
    Check is a string is black or not, either none or only contains whitespace.
    @param string: String to be checked
    @return: Is blank or not
    """
    return string is None or len(string) == 0 or string.isspace()

def metric_to_line_data(name, value, timestamp, source, tags, default_source):
    """Metric Data to String.
    Wavefront Metrics Data format
    <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
    Example: "new-york.power.usage 42422 1533531013 source=localhost
              datacenter=dc1"
    @param name: Metric Name
    @type name: str
    @param value: Metric Value
    @type value: float
    @param timestamp: Timestamp
    @type timestamp: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: dict
    @param default_source:
    @type default_source: str
    @return: String
    """
    if is_blank(name):
        raise ValueError('Metrics name cannot be blank')

    if is_blank(source):
        source = default_source

    str_builder = [sanitize(name), str(float(value))]
    if timestamp is not None:
        str_builder.append(str(int(timestamp)))
    str_builder.append('source=' + sanitize(source))
    if tags is not None:
        for key, val in tags.items():
            if is_blank(key):
                raise ValueError('Metric point tag key cannot be blank')
            if is_blank(val):
                raise ValueError('Metric point tag value cannot be blank')
            str_builder.append(sanitize(key) + '=' + sanitize(val))
    return ' '.join(str_builder) + '\n'

# pylint: disable=unused-argument
def tracing_span_to_line_data(name, start_millis, duration_millis, source,
                              trace_id, span_id, parents, follows_from, tags,
                              span_logs, default_source):
    """Wavefront Tracing Span Data format.
    <tracingSpanName> source=<source> [pointTags] <start_millis>
    <duration_milli_seconds>
    Example: "getAllUsers source=localhost
              traceId=7b3bf470-9456-11e8-9eb6-529269fb1459
              spanId=0313bafe-9457-11e8-9eb6-529269fb1459
              parent=2f64e538-9457-11e8-9eb6-529269fb1459
              application=Wavefront http.method=GET
              1533531013 343500"
    @param name: Span Name
    @type name: str
    @param start_millis: Start time
    @type start_millis: long
    @param duration_millis: Duration time
    @type duration_millis: long
    @param source: Source
    @type source: str
    @param trace_id: Trace ID
    @type trace_id: UUID
    @param span_id: Span ID
    @type span_id: UUID
    @param parents: Parents Span ID
    @type parents: List of UUID
    @param follows_from: Follows Span ID
    @type follows_from: List of UUID
    @param tags: Tags
    @type tags: list
    @param span_logs: Span Log
    @param default_source: Default Source
    @type default_source: str
    @return: String data of tracing span
    """
    # pylint: disable=too-many-locals
    if is_blank(name):
        raise ValueError('Span name cannot be blank')

    if is_blank(source):
        source = default_source

    str_builder = [sanitize(name),
                   'source=' + sanitize(source),
                   'traceId=' + str(trace_id),
                   'spanId=' + str(span_id)]
    if parents is not None:
        for uuid in parents:
            str_builder.append('parent=' + str(uuid))
    if follows_from is not None:
        for uuid in follows_from:
            str_builder.append('followsFrom=' + str(uuid))
    if span_logs:
        tags.append((SPAN_LOG_KEY, 'true'))
    if tags is not None:
        tag_set = set()
        for key, val in tags:
            if is_blank(key):
                raise ValueError('Span tag key cannot be blank')
            if is_blank(val):
                raise ValueError('Span tag val cannot be blank')
            cur_tag = sanitize(key) + '=' + sanitize(val)
            if cur_tag not in tag_set:
                str_builder.append(cur_tag)
                tag_set.add(cur_tag)
    str_builder.append(str(start_millis))
    str_builder.append(str(duration_millis))
    return ' '.join(str_builder) + '\n'

class PyObject(ctypes.Structure):
    _fields_ = [("refcnt", ctypes.c_long)]

CONN_COUNT = 'conn_count'
SENT_COUNT = 'data_sent_count'
SEND_ERROR_SOCKET = 'send_error_socket'
SEND_ERROR_OTHER = 'send_error_other'

class WavefrontProxyConn:
    def __init__(self, address, port,
                 timeout=None):
        """Construct ProxyConnectionHandler.
        @param address: Proxy Address
        @param port: Proxy Port
        @param timeout: Timeout to set on socket connecting to proxy
        """
        self._logger = sha.config.get_logger(__name__)
        self._address = address
        self._port = int(port)
        self.timeout = timeout
        self._reconnecting_socket = None
        self._stats = {
            CONN_COUNT: 0,
            SENT_COUNT: 0,
            SEND_ERROR_SOCKET: 0,
            SEND_ERROR_OTHER: 0,
        }


    def connect(self):
        """Initialize socket and connect to given address:port."""
        self._reconnecting_socket = socket.socket(socket.AF_INET,
                                                  socket.SOCK_STREAM)
        self._reconnecting_socket.settimeout(self.timeout)
        self._reconnecting_socket.connect((self._address, self._port))
        self._stats[CONN_COUNT] += 1

    def close(self):
        """Close socket if it's open now."""
        if self._reconnecting_socket:
            self._logger.debug("{}, Refcount: {}".format(self._reconnecting_socket, sys.getrefcount(self._reconnecting_socket)))
            socket_addr = id(self._reconnecting_socket)
            self._reconnecting_socket.close()
            del self._reconnecting_socket
            self._logger.debug("Refcount after delete: {}".format(PyObject.from_address(socket_addr).refcnt))
            self._reconnecting_socket = None

    def send_data(self, line_data, reconnect=True):
        """Send data via proxy.

        @param line_data: Data to be sent
        @param reconnect: If it's the second time trying to send data
        """
        try:
            if not self._reconnecting_socket:
                self.connect()
            self._reconnecting_socket.sendall(line_data.encode('utf-8'))
            self._stats[SENT_COUNT] += 1
        except socket.error as error:
            self._stats[SEND_ERROR_SOCKET] += 1
            if reconnect:
                # Clear the old socket
                self.close()
                # Try to resend
                self.send_data(line_data, reconnect=False)
            else:
                # Second time trying failed
                raise error
        except Exception as e:
            self._stats[SEND_ERROR_OTHER] += 1
            raise e

    def stats(self):
        return self._stats

METRIC_LIST_COUNT = 'metric_list_count'
METRIC_COUNT = 'metric_count'
METRIC_SENT_COUNT = 'metric_sent_count'
METRIC_ERROR_NO_CONN = 'metric_error_no_conn'
SPAN_LIST_COUNT = 'span_list_count'
SPAN_COUNT = 'span_count'
SPAN_SENT_COUNT = 'span_send_count'
SPAN_ERROR_NO_CONN = 'span_error_no_conn'

class WavefrontProxyClient:
    def __init__(self, host, metrics_port, distribution_port=None, tracing_port=None,
                 timeout=None):
        """Construct wavefrontProxyClient.
        @param host: Proxy Host
        @param metrics_port: Proxy Metric Port
        @param distribution_port: Proxy Distribution Port, not implemented yet
        @param tracing_port: Proxy Tracing Port, not implemented yet
        @param timeout: Timeout to set on socket connecting to proxy
        """
        self._logger = sha.config.get_logger(__name__)
        self._host = host
        self._metrics_port = metrics_port
        self._distribution_port = distribution_port
        self._tracing_port = tracing_port
        self.timeout = timeout

        self._metrics_conn = (
            None if metrics_port is None
            else WavefrontProxyConn(host, metrics_port, timeout=timeout))

        self._tracing_conn = (
            None if tracing_port is None
            else WavefrontProxyConn(host, tracing_port, timeout=timeout))

        self._stats = {
            METRIC_LIST_COUNT: 0,
            METRIC_COUNT: 0,
            METRIC_SENT_COUNT: 0,
            METRIC_ERROR_NO_CONN: 0,
            SPAN_LIST_COUNT: 0,
            SPAN_COUNT: 0,
            SPAN_SENT_COUNT: 0,
            SPAN_ERROR_NO_CONN: 0,
        }

    def close(self):
        if self._metrics_conn:
            self._metrics_conn.close()
            self._logger.info("Wavefront metrics conn stats: {}".format(self._metrics_conn.stats()))
        if self._tracing_conn:
            self._tracing_conn.close()
            self._logger.info("Wavefront tracing conn stats: {}".format(self._tracing_conn.stats()))
        self._logger.info("Wavefront proxy client stats: {}".format(self._stats))

    def send_metric_now(self, metrics):
        if not self._metrics_conn:
            self._logger.warning("Can't send metric data to Wavefront. Please configure "
                                 "metrics port for Wavefront proxy.")
            self._stats[METRIC_ERROR_NO_CONN] += 1
            return

        self._stats[METRIC_LIST_COUNT] += 1
        for metric in metrics:
            self._stats[METRIC_COUNT] += 1
            self._metrics_conn.send_data(metric)
            self._stats[METRIC_SENT_COUNT] += 1

    def send_span_now(self, spans):
        """Send a list of spans immediately.
        Have to construct the data manually by calling
        common.utils.tracing_span_to_line_data()
        @param spans: List of string tracing span data
        @type spans: list[str]
        """
        if not self._tracing_conn:
            self._logger.warning("Can't send tracing span to Wavefront. Please configure "
                                 "tracing port for Wavefront proxy.")
            self._stats[SPAN_ERROR_NO_CONN] += 1
            return

        self._stats[SPAN_LIST_COUNT] += 1
        for span in spans:
            self._stats[SPAN_COUNT] += 1
            self._tracing_conn.send_data(span)
            self._stats[SPAN_SENT_COUNT] += 1
