# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential
from sha.core.plugin.plugin_base import AbstractPlugin
from threading import Lock


class CustomPlugin(AbstractPlugin):
    def __init__(self, id_=None, name=None, type_=None, script=None, manifest=None, config=None, profile_msg_ids=None):
        super(CustomPlugin, self).__init__()
        self._script = script
        self._manifest = manifest
        self._config = config
        self._id = id_
        self._name = name
        self._type = type_
        self._ref_cnt = 1
        self._ref_cnt_lock = Lock()
        self._profile_msg_ids = set()
        if profile_msg_ids is not None:
            self._profile_msg_ids.update(profile_msg_ids)

    def __repr__(self):
        return 'CustomPlugin - id: {}, name:{}, ref_cnt:{}, script: {}, manifest: {}, config: {}, type:{}'.format(
            self._id, self._name, self.ref_cnt, self._script, self._manifest, self._config, self._type)

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def script(self):
        return self._script

    @property
    def manifest(self):
        return self._manifest

    @property
    def config(self):
        return self._config

    @property
    def type(self):
        return self._type

    @property
    def ref_cnt(self):
        """
        reference count of current plugin.
        For the case - multi same plugins installed via different NSGroups, MP cannot block that.
        So, on host, if N same plugins (with same ID) installed, the ref_cnt of that plugin should be N.
        Host can delete plugin only if its ref_cnt = 0
        """
        return self._ref_cnt

    def inc_and_get_cnt(self):
        with self._ref_cnt_lock:
            self._ref_cnt += 1
            return self._ref_cnt

    def dec_and_get_cnt(self):
        with self._ref_cnt_lock:
            self._ref_cnt -= 1
            return self._ref_cnt

    @property
    def profile_msg_ids(self):
        """
        Record this plugin installed by which profile_msgs (1 ServiceConfig mappings to 1 profile_msg). SHA increase
        a plugin's ref_cnt when it's installed multi-times via different profile_msgs (ServiceConfigs).

        But SHA should not repeatedly install the plugin which comes repeatedly from a same profile_msg. Here are some
        examples for this case:
            1. Plugin's ServiceConfig is updated.
            2. SHA full sync with NestDb after reconnection
        :return:
        """
        return self._profile_msg_ids

    def add_msg_id(self, id_):
        self._profile_msg_ids.add(id_)
        return self._profile_msg_ids

    def remove_msg_id(self, id_):
        self._profile_msg_ids.remove(id_)
        return self._profile_msg_ids
