# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import sys
import traceback

import sha.config
import yaml
from sha.core.plugin.custom_plugin_cache import CustomPluginCache
from sha.core.plugin.custom_plugin import CustomPlugin
from sha.core.plugin.custom_plugin_notifier_manager import CustomPluginNotifierManager
from threading import Lock
from sha.config import VERSION
import re
from sha.config import MAX_INSTALLED_CUSTOM_PLUGIN_NUM, NODE_TYPE, PRE_CONDITION, ENABLED_METRICS, EXISTING_METRICS
from sha.constants import VERSION_STR, NODE_TYPE_STR, PRE_CONDITION_STR, MODULES, CLASSES, \
    INSTALL_FAIL, ENABLE, CHECK_INTERVAL, METRICS_STR
from sha.core.plugin.plugin_status_collection import PluginStatusCollection
from sha.core.accesscontrol._module import RestrictedModuleLoader
from sha.utils import convert_to_uuid_str


class CustomPluginResolver(object):
    """
    Resolve custom plugin msg from nestdb.
    """
    _logger = sha.config.get_logger(__name__)
    _custom_resolver_lock = Lock()

    # TODO for mock data - to remove later
    # def __init__(self):
        # self._logger.info("hhtest - load_mock_data start")
        # self.load_mock_data()
        # self._logger.info("hhtest - load_mock_data stop")


    @classmethod
    def resolve(cls, profile_msg, is_clear):
        with cls._custom_resolver_lock:
            """
            Synchronize thread in case multiple nestdb msg concurrently comes leading to data corruption.
            (Not sure about nestdb behavior that whether there will be msg coming concurrently. This lock is added
            just in case. We may confirm it later.)
            """
            cls._logger.info('resolve - is_clear:{}, profile_msg:{}'.format(is_clear, profile_msg))
            if is_clear:
                cls._remove(profile_msg)
            else:
                cls._add(profile_msg)

    @classmethod
    def _remove(cls, profile_msg):
        """
        Handling plugin remove msg.
        Remove plugin from CustomPluginNotifierManager and CustomPluginCache
        """
        plugin_id = profile_msg.system_health_plugin_profile.id
        profile_msg_id = convert_to_uuid_str(profile_msg.id)

        if not cls._should_remove(plugin_id, profile_msg_id):
            return
        CustomPluginNotifierManager.remove_notifier(plugin_id)
        CustomPluginCache.remove_plugin(plugin_id)
        PluginStatusCollection.remove_plugin_status(plugin_id)

    @classmethod
    def _add(cls, profile_msg):
        """
        Handling plugin add msg.
        Construct custom plugin and add it to cache; Add corresponding notifier.
        """
        profile_msg_id = convert_to_uuid_str(profile_msg.id)
        plugin_pb2 = profile_msg.system_health_plugin_profile
        plugin_name = plugin_pb2.name
        plugin_type = plugin_pb2.type
        plugin_id = plugin_pb2.id
        plugin_manifest = plugin_pb2.plugin.manifest
        plugin_script = plugin_pb2.plugin.script
        plugin_config = plugin_pb2.plugin.config

        new_plugin = CustomPlugin(id_=plugin_id, name=plugin_name,
                                  type_=plugin_type, script=plugin_script,
                                  manifest=plugin_manifest, config=plugin_config, profile_msg_ids={profile_msg_id})

        PluginStatusCollection.add_plugin_status(new_plugin.id, new_plugin.name)
        if not cls._should_add(new_plugin):
            return

        CustomPluginCache.add_plugin(new_plugin)
        CustomPluginNotifierManager.add_notifier(new_plugin)

    @classmethod
    def _should_add(cls, new_plugin):
        """
        validation - check whether the plugin should be installed.
        """
        p_id = new_plugin.id
        existed_plugin = CustomPluginCache.get_plugin(p_id)
        if existed_plugin is not None:
            p_msg_id = next(iter(new_plugin.profile_msg_ids))
            if p_msg_id in existed_plugin.profile_msg_ids:
                cls._logger.info('Plugin is already existed for same profile msg. No need to perform installation. '
                                 'plugin_id:{}, profile_msg_id:{}'.format(p_id, p_msg_id))
                return False

            profile_msg_ids = existed_plugin.add_msg_id(p_msg_id)
            ref_cnt = existed_plugin.inc_and_get_cnt()
            cls._logger.info('Plugin is already existed. Reference count is increased. No need to perform installation.'
                             ' plugin_id:{}, ref_cnt:{}, profile_msg_ids:{}'.format(p_id, ref_cnt, profile_msg_ids))
            return False

        if CustomPluginCache.plugin_num() == MAX_INSTALLED_CUSTOM_PLUGIN_NUM:
            err_msg = 'Installed custom plugin count reached max limit:{}'.format(MAX_INSTALLED_CUSTOM_PLUGIN_NUM)
            cls._logger.warn('Failed in custom plugin installation for plugin id:{} '.format(p_id) + err_msg)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            return False

        if not new_plugin.manifest or not new_plugin.script or not new_plugin.config:
            err_msg = 'Plugin data is invalid, some file is empty, , manifest:{}, script:{}, config:{}' \
                .format(new_plugin.manifest, new_plugin.script, new_plugin.config)
            cls._logger.warn('Failed in custom plugin installation, plugin_id:{} '.format(p_id) + err_msg)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            return False
        cls._logger.info('file size, plugin_id:{}, manifest:{}KB, script:{}KB, config:{}KB'
                         .format(p_id, sys.getsizeof(new_plugin.manifest) / 1000,
                                 sys.getsizeof(new_plugin.script) / 1000, sys.getsizeof(new_plugin.config) / 1000))

        try:
            config = yaml.load(new_plugin.config, Loader=yaml.Loader)
            if ENABLE not in config or CHECK_INTERVAL not in config:
                err_msg = 'Invalid config, {} and {} are required. config:{}' \
                    .format(ENABLE, CHECK_INTERVAL, config)
                PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                cls._logger.error('Failed in custom plugin installation, plugin_id:{},{}'
                                  .format(p_id, err_msg))
                return False
        except Exception as e:
            err_msg = 'Invalid config:{}, error:{}'.format(new_plugin.config, e)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            t = traceback.format_exc()
            cls._logger.error('Failed in custom plugin installation, plugin_id:{}, {}, trace:{}'
                              .format(p_id, err_msg, t))
            return False

        try:
            manifest = yaml.load(new_plugin.manifest, Loader=yaml.Loader)
            if VERSION_STR not in manifest or NODE_TYPE_STR not in manifest:
                err_msg = 'Invalid manifest, {} and {} are required. manifest:{}' \
                    .format(VERSION_STR, NODE_TYPE_STR, manifest)
                PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                cls._logger.error('Failed in custom plugin installation, plugin_id:{},{}'
                                  .format(p_id, err_msg))
                return False
        except Exception as e:
            err_msg = 'Invalid manifest:{}, error:{}'.format(new_plugin.manifest, e)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            t = traceback.format_exc()
            cls._logger.error('Failed in custom plugin installation, plugin_id:{}, {}, trace:{}'
                              .format(p_id, err_msg, t))
            return False

        if not cls._match_version(manifest[VERSION_STR]):
            err_msg = 'Version is not match, plugin_version:{}, nsx_version_on_host:{}' \
                .format(manifest[VERSION_STR], VERSION)
            cls._logger.warn('Failed in custom plugin installation, plugin_id:{} '.format(p_id) + err_msg)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            return False

        if NODE_TYPE not in manifest[NODE_TYPE_STR]:
            err_msg = 'Node type is not match, plugin applicable node type:{}, current node type:{}' \
                .format(manifest[NODE_TYPE_STR], NODE_TYPE)
            cls._logger.warn('Failed in custom plugin installation, plugin_id:{}'.format(p_id) + err_msg)
            PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
            return False

        if PRE_CONDITION_STR in manifest:
            preconditions = manifest[PRE_CONDITION_STR]
            for pc in preconditions:
                if pc not in PRE_CONDITION:
                    err_msg = 'Invalid pre-condition:{}. Available pre-conditions:{}'.format(pc, list(PRE_CONDITION.keys()))
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. '.format(p_id) + err_msg)
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

        if METRICS_STR in manifest:
            metrics = manifest[METRICS_STR]
            for m in metrics:
                if m not in EXISTING_METRICS:
                    err_msg = 'Invalid metric:{}. Available metrics:{}.'.format(m, list(EXISTING_METRICS))
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. '.format(p_id) + err_msg)
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

        if MODULES in manifest:
            modules = manifest[MODULES]
            for m in modules:
                if not RestrictedModuleLoader.is_valid_module_name(m):
                    err_msg = "Invalid module name: '{}'".format(m)
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. {}'.format(p_id, err_msg))
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

                if RestrictedModuleLoader.is_forbidden_module(m):
                    err_msg = "Module:'{}' is forbidden to use. Forbidden modules:{}" \
                        .format(m, list(RestrictedModuleLoader.forbidden_modules()))
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. {}'.format(p_id, err_msg))
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

        if CLASSES in manifest:
            classes = manifest[CLASSES]
            for c in classes:
                if not RestrictedModuleLoader.is_valid_class_name(c):
                    err_msg = "Invalid class name: '{}'".format(c)
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. {}'.format(p_id, err_msg))
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

                if RestrictedModuleLoader.is_class_from_forbidden_module(c):
                    err_msg = "Module of class:'{}' is forbidden to use. Forbidden modules:{}" \
                        .format(c, list(RestrictedModuleLoader.forbidden_modules()))
                    cls._logger.warn('Failed in custom plugin installation, plugin_id:{}. {}'.format(p_id, err_msg))
                    PluginStatusCollection.update_plugin_status(p_id, INSTALL_FAIL, err_msg)
                    return False

        return True

    @classmethod
    def _should_remove(cls, plugin_id, profile_msg_id):
        """
        check whether should perform removal.
        """
        current_plugin = CustomPluginCache.get_plugin(plugin_id)
        if current_plugin is None:
            cls._logger.info('Not exist plugin id: {}, no need to perform removal.'.format(plugin_id))
            PluginStatusCollection.remove_plugin_status(plugin_id)
            return False

        profile_msg_ids = current_plugin.remove_msg_id(profile_msg_id)
        ref_cnt = current_plugin.dec_and_get_cnt()
        if ref_cnt > 0:
            cls._logger.info('Plugin is still in use. Reference count is decreased. No need to perform removal. '
                             'plugin_id:{}, ref_cnt:{}, profile_msg_ids:{}'.format(plugin_id, ref_cnt, profile_msg_ids))
            return False
        return True

    @classmethod
    def _match_version(cls, plugin_version):
        pattern = re.compile(plugin_version)
        return pattern.match(VERSION)


    # TODO for mock data - remove later
    #
    # @classmethod
    # def load_mock_data(cls):
    #     import os
    #     plugin_path = os.path.join(os.path.dirname(__file__), "edge", "cp.py")
    #     manifest_path = os.path.join(os.path.dirname(__file__), "edge", "cp.manifest.yml")
    #     default_profile_path = os.path.join(os.path.dirname(__file__), "edge", "cp.yml")
    #     plugin_script = CustomPluginResolver._read_file(plugin_path)
    #     plugin_manifest = CustomPluginResolver._read_file(manifest_path)
    #     plugin_config = CustomPluginResolver._read_file(default_profile_path)
    #     new_plugin = CustomPlugin(id_="fc38cc50-a107-406d-b65c-6fce9011797e", name="hh-test",
    #                               type_="hh-type", script=plugin_script,
    #                               manifest=plugin_manifest, config=plugin_config)
    #
    #     PluginStatusCollection.add_plugin_status(new_plugin.id, new_plugin.name)
    #     if not cls._should_add(new_plugin):
    #         return
    #
    #     CustomPluginCache.add_plugin(new_plugin)
    #     CustomPluginNotifierManager.add_notifier(new_plugin)
    #
    # @staticmethod
    # def _read_file(path):
    #     with open(path, 'r') as f:
    #         try:
    #             file = f.read()
    #             CustomPluginResolver._logger.info("file:{}".format(file))
    #             # content = yaml.load(file, Loader=yaml.Loader)
    #             # cls._logger.info("content:{}".format(content))
    #             return file
    #         except Exception as e:
    #             CustomPluginResolver._logger.info("input_path:{}, SyntaxError:{}", file, e)
