# ***************************************************************************
# Copyright 2017-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import errno
from hashlib import sha256
import json
import os
import socket
import ssl
from subprocess import PIPE
from subprocess import Popen
import sys
import traceback

sys.path.append("/opt/vmware/nsx-common/python")

import nsx_logging


# ###########################################################################
# IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT
# ###########################################################################
#
# If you are updating this file, keep in mind that this file is used by NSX
# services running on hosts (e.g. ESX, RHEL, Ubuntu) registered with the MP.
# This means that Python 3 may be used; therefore, your code has to be both
# Python 2 & 3 compatible.
#
# ###########################################################################
def invoke_command(cmd_list, stdin=None, env=None):
    """Invoke specified cmd_list with optional stdin and env

    Returns:
        (returncode, stdout, stderr) tuple where stdout and stderr values
        have been normalized to strings (i.e. account for Python3
        subprocess.communicate returning byte strings).
    """
    if sys.version_info >= (3,) and stdin is not None:
        # In Python3, communicate() expects stdin as a byte string
        stdin = bytes(stdin, "utf8")
    po = Popen(
        cmd_list, stdin=(PIPE if stdin else None), stdout=PIPE, stderr=PIPE,
        env=env)
    out, err = po.communicate(input=stdin)
    if sys.version_info >= (3,):
        # In Python3, communicate() returns out and err as byte strings
        if out is not None:
            out = out.decode("utf8")
        if err is not None:
            err = err.decode("utf8")
    return po.returncode, out, err

if sys.version_info < (3,):
    from cStringIO import StringIO
    import httplib
    _HTTPLIB_OK = httplib.OK
    _HTTPLIB_BAD_REQUEST = httplib.BAD_REQUEST
    _HTTPLIB_SERVICE_UNAVAILABLE = httplib.SERVICE_UNAVAILABLE
    from httplib import HTTPSConnection
else:
    from io import StringIO
    import http.client
    _HTTPLIB_OK = http.client.OK
    _HTTPLIB_BAD_REQUEST = http.client.BAD_REQUEST
    _HTTPLIB_SERVICE_UNAVAILABLE = http.client.SERVICE_UNAVAILABLE
    from http.client import HTTPSConnection


UTILS_DIR = "/opt/vmware/nsx-platform-client/bin/python"


def _upload_to_manager(input_obj):
    manager_info = input_obj["manager_info"]
    tmp_log_path = input_obj["local_path"]
    try:
        # Suppress default Python 2.7.9+ HTTPSConnection certificate and
        # hostname verification. Doing manual certificate thumbprint check
        # instead.
        if sys.version_info >= (2, 7, 9):
            ctx = ssl._create_unverified_context()
            conn = HTTPSConnection(
                manager_info["server"], manager_info["port"], context=ctx,
                timeout=30)
        else:
            conn = HTTPSConnection(
                manager_info["server"], manager_info["port"], timeout=30)
        conn.connect()
    except socket.error as ex:
        lg.warning(
            "Unable to connect to manager at %s:%s" %
            (manager_info["server"], manager_info["port"]))
        sys.stderr.write(
            "Unable to connect to manager at %s:%s\n" %
            (manager_info["server"], manager_info["port"]))
        return 1
    try:
        # Verify server thumbprint
        conn_thumbprint = sha256(conn.sock.getpeercert(True)).hexdigest()
        conn_thumbprint = conn_thumbprint.lower()
        if conn_thumbprint != manager_info["thumbprint"]:
            lg.warning("Manager thumbprint mismatch")
            sys.stderr.write("Manager thumbprint mismatch\n")
            return 1
        size = os.stat(tmp_log_path).st_size
        headers = {
            "Content-Length": "%s" % size,
            "Content-Type": "application/octet-stream",
            "X-Nsx-Content-Length": "%s" % size}
        path = "/api/v1/administration/audit-logs/%s" % (
            input_obj["generate_id"])
        with open(tmp_log_path, "rb") as fo:
            conn.request("POST", path, fo, headers)
        resp = conn.getresponse()
        resp_body = resp.read()
        if resp.status == _HTTPLIB_OK:
            sys.stdout.write("MANAGER&%s" % tmp_log_path)
            return 0
        elif resp.status == _HTTPLIB_SERVICE_UNAVAILABLE:
            lg.warning(
                "NSX Manager has insufficient space to receive the audit log")
            sys.stderr.write(
                "NSX Manager has insufficient space to receive the audit log\n")
        elif resp.status == _HTTPLIB_BAD_REQUEST:
            lg.warning(
                "NSX Manager has rejected the upload request: %s" % resp_body)
            sys.stderr.write(
                "NSX Manager has rejected the upload request: %s\n" % resp_body)
        else:
            lg.warning(
                "NSX Manager reject request, status: %s, body: %s" %
                (resp.status, resp_body))
            sys.stderr.write(
                "NSX Manager reject request, status: %s, body: %s\n" %
                (resp.status, resp_body))
    except Exception:
        lg.warning(
            "Unable to upload file to manager: %s" % traceback.format_exc())
        sys.stderr.write(
            "Unable to upload file to manager: %s\n" % traceback.format_exc())
    finally:
        if conn:
            conn.close()
        # Deleted the collected local audit logs in /tmp
        try:
            lg.info("Deleting log file %s", tmp_log_path)
            os.remove(tmp_log_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                lg.warning("Failed to delete %s: %s", tmp_log_path,
                    traceback.format_exc())
            pass
    return 1


def main(argv):
    # The caller will send send json to stdin in the form of:
    # {
    #   "local_path": string,
    #   "generate_id": string,
    #   "manager_info": {
    #     "port": integer,
    #     "server": string,
    #     "thumbprint": string}}
    #
    # If upload successful, exit status is 0, "MANAGER" will have
    # been written to stdout (where MANAGER" indicates log
    # successfully uploaded fo file server/manager respectively.
    global lg
    nsx_logging.basicConfig(syslog=True, subcomp="audit-upload")
    lg = nsx_logging.getLogger(__name__)
    lg.setLevel(nsx_logging.INFO)
    lg.info("Start uploading audit logs")
    input_obj = json.loads(sys.stdin.read())
    return _upload_to_manager(input_obj)


if __name__ == "__main__":
    sys.exit(main(sys.argv))
