#!/usr/bin/env python

import cmd
import functools
import gevent
import inspect
import interface
import re
import sys

from vmware.nsx.rpc import NsxRpcClient


"""Decorator to provide rpc connection to each command call
Takes in a instance of Interface, it will use interface.stub and interface.endpoint
to establish RPC connection before each command call, and destroy the connection
after. The reason of doing this is to avoid connection timeout."""
def requireNsxRpcClient(interface):
    def inner_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(5):
                try:
                    interface.client = NsxRpcClient([interface.stub], interface.endpoint)
                    break
                except Exception as e:
                    print("trying to connect to server ...")
                    gevent.sleep(3)
            assert interface.client, "client failed to connect"
            interface.client.connection.keep_alive_timeout_sec = 3600
            try:
                ret = func(*args, **kwargs) # actual function call
                return ret
            except Exception as e:
                sys.stderr.write("ProxyCliClient encountered an error: %s\n" % (str(e)))
            finally:
                interface.client.Destroy()
        return wrapper
    return inner_decorator


class CliTool(cmd.Cmd):
    """ Generates a single nesting of CLIs from each of the service-associated
    interface classes. User types the name of the service they want to use, then
    can call specific methods from that service. Takes in a list of desired
    interfaces and a server listening endpoint. """

    intro = "A command line tool for getting info about TN proxy services. Type 'help' to see list of available commands."
    ruler = '-'
    prompt = '(tn-proxy-cli) '
    doc_header = 'System commands:'
    undoc_header = 'Available interfaces:'

    def __init__(self, interfaces, endpoint):
        cmd.Cmd.__init__(self)
        self.endpoint = endpoint
        self.interfaces = interfaces
        for i in self.interfaces:
            i.endpoint = self.endpoint

        self.subtools = {}

        # create subtool objects and generate do_subtool commands
        self.make_subtools()
        for subtool in self.subtools:
            self.make_do_subtool(subtool)

    def make_subtools(self):
        """ Creates a CliSubTool for each interface and generates a
        corresponding 'do_subtool' function. """

        for interface in self.interfaces:
            name = interface.__class__.__name__
            subtool = CliSubTool(interface, name)
            self.subtools[name] = subtool

    def make_do_subtool(self, name):
        fn_name = "do_" + name

        def fn(self, name):
            tool = self.subtools[fn.__name__]
            tool.cmdloop()

        setattr(CliTool, fn_name, fn)
        fn.__name__ = name

    def do_quit(self, quit):
        print('')
        return True

    def help_quit(self):
        print('Exit the CLI.')

class CliSubTool(cmd.Cmd):
    """ CliSubTool creates a CLI corresponding to a single interface. Generates
    command methods corresponding to interface request methods. Expects an
    interface instance and interface name passed from CliTool.

    The Cmd module is designed to display class methods which begin with 'do_'
    to the user. Since multiple CliSubTool instances will be present and
    managed by CliTool at once, we bind generated methods to CliSubTool
    instances and override Cmd's get_names() method so that each CliSubTool
    will only display its own commands to the user."""

    def __init__(self, interface, name):
        cmd.Cmd.__init__(self)
        self.name = name
        self.interface = interface
        self.commands = []

        self.prompt = '(tn-proxy-cli:' + self.name + ') '
        self.intro = "\n" + self.name + ": type 'help' for list of available " + \
        "commands, or 'help CommandName' for information about a command.\nType " + \
        "'quit' to return to main CLI.\n"
        self.ruler = '-'
        self.doc_header = 'System commands:'
        self.undoc_header = 'Interface commands:'

        self.make_get_commands()

    # Overrides the get_names method in the Cmd module so that the CLI will
    # display generated methods bound to CliSubTool instances.
    def get_names(self):
        return dir(self.__class__) + self.commands

    def make_get_commands(self):
        """Generate commands for each 'get_' function from interfaces."""

        methods = inspect.getmembers(self.interface)
        for m in methods:
            # get 'command' from function name
            if re.match(r"get.*", m[0]):
                command = m[0].split("get_", 1)[1]
                self.make_command(command)
            if re.match(r"help.*", m[0]):
                command = m[0].split("help_", 1)[1]
                self.make_help(command)

    # generate the 'do_' commands recognized by the cmd module
    def make_command(self, command):
        fn_name = "do_" + command

        @requireNsxRpcClient(self.interface)
        def fn(args):
            request = "get_" + fn.__name__
            args = args.split()
            print(getattr(self.interface, request)(args))

        setattr(self, fn_name, fn)
        fn.__name__ = command
        self.commands.append(fn_name)

    def make_help(self, command):
        fn_name = "help_" + command

        def fn():
            rsp =  getattr(self.interface, fn_name)()
            print("")
        setattr(self, fn_name, fn)

    def do_quit(self, quit):
        print('')
        return True

    def help_quit(self):
        print('Exit the current interface and return to main CLI.')

